/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 * Copyright 1993 David B. Hollenbeck
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

/*
 * DlgKit definitions
 */

#include <Dlg/dlgdb.h>
#include <Dlg/dlgcov.h>
#include <Dlg/dlgfclass.h>
#include <Dlg/dlglib.h>

#include <DlgUnidraw/dlgkit.h>
#include <DlgUnidraw/dlgcmds.h>
#include <DlgUnidraw/dlgcomps.h>
#include <DlgUnidraw/dlgeditor.h>

#include <MapUnidraw/gscmd.h>
#include <MapUnidraw/mapabout.h>
#include <MapUnidraw/mapcamera.h>
#include <MapUnidraw/mapcmds.h>
#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapexport.h>
#include <MapUnidraw/mapinspect.h>
#include <MapUnidraw/mappoint.h>
#include <MapUnidraw/mapproj.h>
#include <MapUnidraw/mapstatecmds.h>
#include <MapUnidraw/mapstates.h>

#include <OverlayUnidraw/ovcmds.h>
#include <OverlayUnidraw/ovcomps.h>
#include <OverlayUnidraw/ovdoer.h>
#include <OverlayUnidraw/oved.h>
#include <OverlayUnidraw/ovimport.h>
#include <OverlayUnidraw/ovprint.h>
#include <OverlayUnidraw/ovviewer.h>

#define iv2_6_compatible 1
#include <UniIdraw/idcmds.h>
#include <UniIdraw/idkybd.h>
#undef iv2_6_compatible

#include <Unidraw/unidraw.h>
#include <Unidraw/editor.h>
#include <Unidraw/ctrlinfo.h>
#include <Unidraw/iterator.h>
#include <Unidraw/uctrls.h>
#include <Unidraw/keymap.h>
#include <Unidraw/kybd.h>
#include <Unidraw/catalog.h>
#include <Unidraw/viewer.h>

#include <Unidraw/Commands/catcmds.h>

#include <Unidraw/Graphic/pspaint.h>

#include <Unidraw/Tools/magnify.h>

#include <IV-X11/xcolor.h>
#include <IV-X11/xdisplay.h>
#include <IV-X11/xpattern.h>
#include <IV-look/kit.h>
#include <InterViews/action.h>
#include <InterViews/background.h>
#include <InterViews/display.h>
#include <InterViews/label.h>
#include <InterViews/layout.h>
#include <InterViews/patch.h>
#include <InterViews/style.h>
#include <InterViews/session.h>

#include <IVGlyph/exportchooser.h>
#include <IVGlyph/figure.h>
#include <IVGlyph/toolbutton.h>

#include <Map/mapdb.h>
#include <Map/maplib.h>
#include <MapUnidraw/mapeditor.h>

#include <OS/math.h>

#include <stdlib.h>
#include <string.h>

implementTable(DlgFClassMenuTable,DlgFeatureClass*,MenuItem*)
implementTable(DlgFClassTable,DlgFeatureClass*,boolean)

/******************************************************************************/

DlgKit* DlgKit::_dlgkit = nil;

DlgKit::DlgKit()
: MapKit()
{
    InitFClassState();
}

DlgKit::~DlgKit() {
    delete _fcmenutable;
    delete _fctable;
}

DlgKit* DlgKit::Instance() {
    if (!_dlgkit)
	_dlgkit = new DlgKit();
    return _dlgkit;
}

Glyph* DlgKit::MakeToolbar() {
    WidgetKit& kit = *WidgetKit::instance();
    const LayoutKit& layout = *LayoutKit::instance();
    Style* s = kit.style();

    ToolButton* magnify;
    ToolButton* mapselect;
    ToolButton* inspect;
#ifdef HAVE_ACE
    ToolButton* camerapos;
    ToolButton* camerapath;
#endif
    _toolbars = layout.deck(2);

    PolyGlyph* vb = layout.vbox(5);

    TelltaleGroup* tg = new TelltaleGroup();

    Glyph* mag = kit.label("Magnify");
    Glyph* mps = kit.label("Select");
    Glyph* ins = kit.label("Inspect");
#ifdef HAVE_ACE
    Glyph* campos = kit.label("CameraPos");
    Glyph* campth = kit.label("CameraPath");
#endif

    Coord maxwidth = 0;
    Requisition req;
    maxwidth = Math::max((mag->request(req), req.x_requirement().natural()),
			 maxwidth);
    maxwidth = Math::max((mps->request(req), req.x_requirement().natural()),
			 maxwidth);
    maxwidth = Math::max((ins->request(req), req.x_requirement().natural()),
			 maxwidth);
#ifdef HAVE_ACE
    maxwidth = Math::max((campos->request(req), req.x_requirement().natural()),
			 maxwidth);
    maxwidth = Math::max((campth->request(req), req.x_requirement().natural()),
			 maxwidth);
#endif

    vb->append(magnify = MakeTool(new MagnifyTool(new ControlInfo("Magnify", KLBL_MAGNIFY,CODE_MAGNIFY)),
			layout.overlay(layout.hcenter(layout.hspace(maxwidth)),
				       layout.hcenter(mag)), tg));
    vb->append(mapselect = MakeTool(new MapSelectTool(new ControlInfo("Select", "s", "s")),
			layout.overlay(layout.hcenter(layout.hspace(maxwidth)),
				       layout.hcenter(mps)), tg));
    vb->append(inspect = MakeTool(new MapInspectTool(((MapEditor*)_ed)->selstate(),
					   new ControlInfo("Inspect", "i", "i")),
			layout.overlay(layout.hcenter(layout.hspace(maxwidth)),
				       layout.hcenter(ins)), tg));
#ifdef HAVE_ACE
    vb->append(camerapos = MakeTool(new CameraPosTool(
	new ControlInfo("CameraPos", "c", "c")),
			layout.overlay(layout.hcenter(layout.hspace(maxwidth)),
						layout.hcenter(campos)), tg));
    vb->append(camerapath = MakeTool(new CameraPathTool(
	new ControlInfo("CameraPath", "C", "C")),
			layout.overlay(layout.hcenter(layout.hspace(maxwidth)),
						layout.hcenter(campth)), tg));
#endif
    _toolbars->append(vb);

    vb = layout.vbox(5);
    vb->append(magnify);
    vb->append(mapselect);
    vb->append(inspect);
#ifdef HAVE_ACE
    vb->append(camerapos);
    vb->append(camerapath);
#endif
    _toolbars->append(vb);
    _toolbars->flip_to(1);
    _toolbar = new Patch(_toolbars);

    return layout.hbox(
	layout.vflexible(
	    layout.vnatural(
		new Background(
		    layout.vcenter(
			_toolbar
		    ),
		    unidraw->GetCatalog()->FindColor("#aaaaaa")
		),
		550
	    )
	)
    );
}

Glyph* DlgKit::MakeMenus()
{
    LayoutKit& lk = *LayoutKit::instance();
    Menu* menubar_ = WidgetKit::instance()->menubar();
    
    menubar_->append_item(MakeFileMenu());
    menubar_->append_item(MakeViewMenu());
    menubar_->append_item(MakeMapMenu());
    Resource::ref(menubar_);
    return LayoutKit::instance()->vbox(lk.hnatural(menubar_, 600));
}

MenuItem * DlgKit::MakeFileMenu() {
    LayoutKit& lk = *LayoutKit::instance();
    WidgetKit& kit = *WidgetKit::instance();
    Style * style = new Style(Session::instance()->style());
    
    MenuItem *mbi = kit.menubar_item(kit.label("File"));
    mbi->menu(kit.pulldown());
    MakeMenu(mbi, new MapAboutCmd(new ControlInfo("About ivmaps dlgviewer", KLBL_ABOUT, CODE_ABOUT)),
	     "About ivmaps dlgviewer");
    MakeMenu(mbi, new OvNewCompCmd(new ControlInfo("New", KLBL_NEWCOMP, CODE_NEWCOMP),
				 new DlgIdrawComp),
	     "New   ");
    MakeMenu(mbi, new OvRevertCmd(new ControlInfo("Revert", KLBL_REVERT, CODE_REVERT)),
	     "Revert   ");
    MakeMenu(mbi, new MapOpenCmd(new ControlInfo("Open...", KLBL_VIEWCOMP, CODE_VIEWCOMP)),
	     "Open...   ");
    MakeMenu(mbi, new OvSaveCompCmd(new ControlInfo("Save", KLBL_SAVECOMP, CODE_SAVECOMP)),
	     "Save   ");
    MakeMenu(mbi, new OvSaveCompAsCmd(new ControlInfo("Save As...",
					    KLBL_SAVECOMPAS,
					    CODE_SAVECOMPAS)),
	     "Save As...   ");
    MakeMenu(mbi, new OvPrintCmd(new ControlInfo("Print...", KLBL_PRINT, CODE_PRINT)),
	     "Print...   ");
    style->attribute("subcaption", "Export selected map features to file:");
    style->attribute("open", "Export");
    const char *formats[] = {"idraw", "drawtool"};
    ExportChooser* echooser = new ExportChooser(".", WidgetKit::instance(), style,
				 formats, sizeof(formats)/sizeof(char*), nil, nil, true, false);
    Resource::ref(echooser);
    MakeMenu(mbi, new MapExportCmd(new ControlInfo("Export Graphic...",
						  "^X", "\030"),
			       echooser),
	     "Export Graphic...   ");
    MakeMenu(mbi, new OvWindowDumpAsCmd(new ControlInfo("Dump Window As..."
						  )),
	     "Dump Window As...   ");
    mbi->menu()->append_item(kit.menu_item_separator());
    MakeMenu(mbi, new OvQuitCmd(new ControlInfo("Quit", KLBL_QUIT, CODE_QUIT)),
	     "Quit   ");

    return mbi;
}

MenuItem* DlgKit::MakeMapMenu() {
    LayoutKit& lk = *LayoutKit::instance();
    WidgetKit& kit = *WidgetKit::instance();
    
    MenuItem *mbi = kit.menubar_item(kit.label("Map"));
    mbi->menu(kit.pulldown());
    // 1
    _libmenu = MakeLibraryMenu();
    mbi->menu()->append_item(_libmenu);

    // 2
    _layermenu = MakeLayerMenus();
    mbi->menu()->append_item(_layermenu);

    _mapmenu = mbi;
    return mbi;
}

void DlgKit::MakeCheckMenu(MenuItem* mbi, Command* scmd, Command* hcmd,
			   const char* label, DlgFeatureClass* fc) {
    LayoutKit& lk = *LayoutKit::instance();
    WidgetKit& kit = *WidgetKit::instance();
    MenuItem* item;
    boolean oldval;
    _ed->GetKeyMap()->Register(new CommandControl(scmd->GetControlInfo()));
    scmd->SetEditor(_ed);
    _ed->GetKeyMap()->Register(new CommandControl(hcmd->GetControlInfo()));
    hcmd->SetEditor(_ed);
    MenuItem* menu_item = kit.check_menu_item(kit.label(label));
    _fcmenutable->insert(fc, menu_item);
    ShowHideDoer* doer = new ShowHideDoer(scmd, hcmd, menu_item->state());
    menu_item->action(new ActionCallback(ShowHideDoer)(doer, &ShowHideDoer::Do));
    if (_fctable->find(oldval, fc) && oldval)
	menu_item->state()->set(TelltaleState::is_chosen, true);
    mbi->menu()->append_item(menu_item);
}

MenuItem* DlgKit::MakeLayerMenus() {
    WidgetKit& kit = *WidgetKit::instance();
    MenuItem* showi = kit.menu_item(kit.label(   "Layers                     "));
    Menu* show = kit.pullright();
    showi->menu(show);
    DlgEditor* ed = (DlgEditor*)GetEditor();
    DlgDatabase* db = ed->GetMapDatabase();
    if (db) {
    DlgLibrary* lib = db->library(ed->libstate()->maplib());
    if (lib) {
	int maxcovlen =0;
	for (int ic = 0; ic < lib->ncovs(); ic++) {
	    DlgCoverage* cov = lib->coverage(ic);
	    int len = strlen(cov->description());
	    if (len > maxcovlen)
		maxcovlen = len;
	}
	for (int ic = 0; ic < lib->ncovs(); ic++) {
	    DlgCoverage* cov = lib->coverage(ic);
	    char covdesc[maxcovlen+1];
	    memset(covdesc, ' ', maxcovlen);
	    strncpy(covdesc, cov->description(), strlen(cov->description()));
	    covdesc[maxcovlen] = '\0';
	    MenuItem* showcovi = kit.menu_item(kit.label(covdesc));
	    Menu* showcovm = kit.pullright();
	    showcovi->menu(showcovm);
	    int maxfclen =0;
	    for (int ifc = 0; ifc < cov->nfcls(); ifc++) {
		DlgFeatureClass* fc = cov->feature_class(ifc);
		int len = strlen(fc->description());
		if (len > maxfclen)
		    maxfclen = len;
	    }
	    for (int ifc = 0; ifc < cov->nfcls(); ifc++) {
		DlgFeatureClass* fc = cov->feature_class(ifc);
		MapShowCmd* showcmd = new MapShowCmd(new ControlInfo(fc->description()), false);
		showcmd->SetLibName(lib->name());
		showcmd->SetCovName(cov->name());
		showcmd->SetFClassName(fc->name());
		MapHideCmd* hidecmd = new MapHideCmd(new ControlInfo(fc->description()), false);
		hidecmd->SetLibName(lib->name());
		hidecmd->SetCovName(cov->name());
		hidecmd->SetFClassName(fc->name());
		char fcdesc[maxfclen+1];
		memset(fcdesc, ' ', maxfclen);
		strncpy(fcdesc, fc->description(), strlen(fc->description()));
		fcdesc[maxfclen] = '\0';
		// if fc shown set check for menu
		MakeCheckMenu(showcovi, showcmd, hidecmd, fcdesc, fc);
	    }
	    show->append_item(showcovi);
	}
    }
    }
    return showi;
}

void DlgKit::FClassState(DlgFeatureClass* fc, boolean val) {
    MenuItem* item;
    boolean oldval;
    if (_fcmenutable->find(item, fc))
	item->state()->set(TelltaleState::is_chosen, val);
    _fctable->insert(fc, val);
}

void DlgKit::InitFClassState() {
    if (_fcmenutable)
        delete _fcmenutable;
    _fcmenutable = new DlgFClassMenuTable(100);
    if (_fctable)
        delete _fctable;
    _fctable = new DlgFClassTable(100);    
}

MenuItem* DlgKit::MakeLibraryMenu() {
    WidgetKit& kit = *WidgetKit::instance();
    MenuItem* libi = kit.menu_item(kit.label(   "Library                   "));
    Menu* libm = kit.pullright();
    libi->menu(libm);
    DlgEditor* ed = (DlgEditor*)GetEditor();
    DlgDatabase* db = ed->GetMapDatabase();
    int maxliblen = 0;
    if (db) {
    for (int il = 0; il < db->nlibs(); il++) {
	DlgLibrary* lib = db->library(il);
	int len = strlen(lib->name());
	if (len > maxliblen)
	    maxliblen = len;
    }
    TelltaleGroup* tg = new TelltaleGroup();
    for (int il = 0; il < db->nlibs(); il++) {
	DlgLibrary* lib = db->library(il);
	char libname[maxliblen+1];
	memset(libname, ' ', maxliblen);
	libname[maxliblen] = '\0';
	strncpy(libname, lib->name(), strlen(lib->name()));
	MenuItem* menu_item = kit.radio_menu_item(tg, libname);
	if (strcmp(lib->name(), ed->libstate()->maplib())==0)
	    menu_item->state()->set(TelltaleState::is_chosen, true);
	MapLibCmd* cmd = new MapLibCmd(new ControlInfo(""), lib->name());
	cmd->SetEditor(ed);
	CommandPusher* pusher = new CommandPusher(cmd);
	menu_item->action(new ActionCallback(CommandPusher)
			  (pusher, &CommandPusher::Push));
	libm->append_item(menu_item);
    }
    }
    return libi;
}

void DlgKit::UpdateLayerMenu() {
    _mapmenu->menu()->replace_item(1, MakeLayerMenus());
}

void DlgKit::UpdateLibraryMenu() {
    _mapmenu->menu()->replace_item(2, MakeLibraryMenu());
}

void DlgKit::InitMembers (OverlayComp* comp) {
  OverlayKit::InitMembers(comp);
}

Glyph* DlgKit::MakeStates() {
    Catalog* catalog = unidraw->GetCatalog();
    DlgDbComp* dbcomp = (DlgDbComp*)((DlgEditor*)_ed)->GetMapDbComp();
    DlgLibComp* libcomp = nil;
    DlgCovComp* covcomp = nil;
    DlgFClassComp* fccomp = nil;
    if (dbcomp) {
      Iterator libit;
      dbcomp->Last(libit);
      libcomp = (DlgLibComp*)dbcomp->GetComp(libit);
      if (libcomp) {
	Iterator covit;
	libcomp->Last(covit);
	covcomp = (DlgCovComp*)libcomp->GetComp(covit);
	if (covcomp) {
	  Iterator fcit;
	  covcomp->Last(fcit);
	  fccomp = (DlgFClassComp*)covcomp->GetComp(fcit);
	}
      }
    }

    ((DlgEditor*)_ed)->_fclassstate = new MapFClassState
      (fccomp ? fccomp->GetFeatureClass()->name() : "",
       fccomp ? fccomp->GetFeatureClass()->description() : "");
    ((DlgEditor*)_ed)->_covstate = new MapCovState
      (((DlgEditor*)_ed)->_fclassstate,
       covcomp ? covcomp->GetCoverage()->name() : "",
       covcomp ? covcomp->GetCoverage()->description() : "");
    ((DlgEditor*)_ed)->_libstate = new MapLibState
      (((DlgEditor*)_ed)->_covstate,
       libcomp ? libcomp->GetLibrary()->name() : "");
    ((DlgEditor*)_ed)->_dbstate = new MapDbState
      (((DlgEditor*)_ed)->_libstate,
       dbcomp->GetDatabase() ? dbcomp->GetDatabase()->name(): 
       (catalog->GetAttribute("database") ? catalog->GetAttribute("database") :
	""));
    
    ((DlgEditor*)_ed)->_projstate = new ProjectionState();
    ((DlgEditor*)_ed)->_selstate = new MapSelectionState("");

    NameView* fclassview = new NameView(((DlgEditor*)_ed)->_fclassstate);
    NameView* covview = new NameView(((DlgEditor*)_ed)->_covstate);
    NameView* libview = new NameView(((DlgEditor*)_ed)->_libstate);
    NameView* dbview = new NameView(((DlgEditor*)_ed)->_dbstate);
    NameView* projview = new NameView(((DlgEditor*)_ed)->_projstate);
    NameView* selview = new NameView(((DlgEditor*)_ed)->_selstate);

    LayoutKit& lk = *LayoutKit::instance();
    WidgetKit& kit = *WidgetKit::instance();
    return kit.outset_frame(
	lk.hmargin(
	    lk.vbox(
		lk.vmargin(lk.hbox(dbview, projview), 2),
		lk.vmargin(libview, 2),
		lk.vmargin(selview, 2)
	    ),
	    4,0,0,0.0,fil,0.0
	)
    );
}
