/*
 * Copyright (c) 1997-1998 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <Attribute/aliterator.h>
#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <MapServ/geomdist.h>
#include <MapServ/mapfunc.h>
#include <MapServ/maphandler.h>
#include <ComTerp/comterpserv.h>
#include <ComTerp/comvalue.h>
#include <Dlg/dlgdb.h>
#include <Dlg/dlglib.h>
#include <Dlg/dlgcov.h>
#include <Dlg/dlgfclass.h>
#include <Vpf/vpfdb.h>
#include <Vpf/vpflib.h>
#include <Vpf/vpfcov.h>
#include <Vpf/vpffclass.h>
#include <TopoFace/fgeomobjs.h>
#include <TopoFace/topoedge.h>
#include <TopoFace/topoface.h>
#include <TopoFace/toponode.h>
#include <Map/mapfeature.h>
#include <Map/mapprim.h>
#if !defined(DLG_FREE)
#include <DlgUnidraw/dlgclasses.h>
#include <DlgUnidraw/dlgcomps.h>
#include <DlgUnidraw/dlgcreator.h>
#endif
#if !defined(VPF_FREE)
#include <VpfUnidraw/vpfclasses.h>
#include <VpfUnidraw/vpfcomps.h>
#include <VpfUnidraw/vpfcreator.h>
#include <VpfUnidraw/vpfscripts.h>
#endif
#include <MapUnidraw/mapcatalog.h>
#include <MapUnidraw/mapclasses.h>
#include <MapUnidraw/mapcmds.h>
#include <MapUnidraw/mapcreator.h>
#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapeditor.h>
#include <MapUnidraw/mapscripts.h>
#include <MapUnidraw/mapviewstate.h>
#include <OverlayUnidraw/clipline.h>
#include <OverlayUnidraw/cliplinepoly.h>
#include <OverlayUnidraw/clippoly.h>
#include <OverlayUnidraw/ovhull.h>
#include <Unidraw/clipboard.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/Graphic/geomobjs.h>
#include <Unidraw/Graphic/graphic.h>
#include <InterViews/transformer.h>
#include <OS/math.h>

#ifdef HAVE_ACE
#include <ace/SOCK_Connector.h>
#endif

extern "C" {
#include <projects.h>
}

#include <strstream.h>
#include <stdio.h>

/*****************************************************************************/

MapFunc::MapFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void MapFunc::insert_array(MapFeature**& array, int& nf, int& cnf,
				MapFeature* f)
{
  if (cnf == nf) {
    MapFeature** newfeats = new MapFeature*[cnf*2];
    for (int j = 0; j < cnf; j++)
      newfeats[j] = array[j];
    delete array;
    array = newfeats;
    nf = cnf * 2;
  }
  array[cnf++] = f;
}

MapFeature* MapFunc::copy_point(MapFeature* f, MapFeatureComp* fcomp) {
  MapFeature* newf = new MapFeature(f->id(), MapFeature::PointType);
  newf->attrlist(new AttributeList(f->attrlist()));
  AttributeList* compal = fcomp->GetAttributeList();
  ALIterator alit;
  for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
    newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
  MapPoint* newp = new MapPoint(newf);
  MapPoint* point = f->point_primitive();
  newp->insert_points(1, (float*)point->xpoints(), (float*)point->ypoints());
  newf->point_primitive(newp);
  newf->graphic((void*)new MapFeatureComp(newf));
  return newf;
}

MapFeature* MapFunc::copy_line(MapFeature* f, MapFeatureComp* fcomp) {
  MapFeature* newf = new MapFeature(f->id(), MapFeature::LineType);
  newf->attrlist(new AttributeList(f->attrlist()));
  AttributeList* compal = fcomp->GetAttributeList();
  ALIterator alit;
  for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
    newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
  MapEdge* newe = new MapEdge(newf);
  MapEdge* edge = f->edge_primitive();
  int npts = edge->npts();
  float* fx = (float*)edge->xpoints();
  float* fy = (float*)edge->ypoints();
  newe->insert_points(npts, (float*)fx, (float*)fy);
  newf->edge_primitive(newe);
  newf->graphic((void*)new MapFeatureComp(newf));
  return newf;
}

MapFeature* MapFunc::copy_area(MapFeature* f, MapFeatureComp* fcomp) {
  MapFeature* newf = new MapFeature(f->id(), MapFeature::AreaType);
  newf->attrlist(new AttributeList(f->attrlist()));
  AttributeList* compal = fcomp->GetAttributeList();
  ALIterator alit;
  for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
    newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
  MapFace* newface = new MapFace(newf);
  MapFace* face = f->face_primitive();
  int npts = face->npts();
  float* fx = (float*)face->xpoints();
  float* fy = (float*)face->ypoints();
  newface->insert_points(npts, (float*)fx, (float*)fy);
  newf->face_primitive(newface);
  newf->graphic((void*)new MapFeatureComp(newf));
  return newf;
}

MapFeature* MapFunc::copy_junction(MapFeature* f, MapFeatureComp* fcomp) {
  MapFeature* newf = new MapFeature(f->id(), MapFeature::JunctionType);
  newf->attrlist(new AttributeList(f->attrlist()));
  AttributeList* compal = fcomp->GetAttributeList();
  ALIterator alit;
  for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
    newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
  MapNode* newp = new MapNode(newf);
  MapNode* node = f->node_primitive();
  newp->insert_points(1, (float*)node->xpoints(), (float*)node->ypoints());
  newf->node_primitive(newp);
  newf->graphic((void*)new MapFeatureComp(newf));
  return newf;
}

/*****************************************************************************/

#if !defined(VPF_FREE)

LoadVpfFCFunc::LoadVpfFCFunc(ComTerp* comterp) : ComFunc(comterp) {
  _tileid_symid = symbol_add("tileid");
}

void LoadVpfFCFunc::execute() {
    ComValue dbpath(stack_arg(0));
    ComValue libname(stack_arg(1));
    ComValue covname(stack_arg(2));
    ComValue fcname(stack_arg(3));
    ComValue tileid(stack_key(_tileid_symid));
    reset_stack();
    ComValue result;

    Creator::instance(new VpfCreator());
    const char* dbdir = dbpath.string_ptr();
    VpfDatabase* db = new VpfDatabase(dbdir);
    VpfLibrary* lib = db->library(libname.string_ptr());
    VpfDbComp* dbcomp = new VpfDbComp(db);
    MapShowCmd* showcmd = new MapShowCmd();
    showcmd->SetUseEditor(false);
    showcmd->SetLibName(libname.string_ptr());
    showcmd->SetCovName(covname.string_ptr());
    showcmd->SetFClassName(fcname.string_ptr());
    VpfExtent* ext = lib->extent();
    char lon0[80];
    sprintf(lon0, "lon_0=%f", ext->x1);  // let proj pick the utm zone
    char* parms[4];
    parms[0] = "proj=utm";
    parms[1] = "ellps=WGS84";
    parms[2] = lon0;
    parms[3] = "no_defs";
    PJ* proj = pj_init(4, parms);
    showcmd->SetProjection(proj);
    if (tileid.type() != ComValue::UnknownType) {
      int tid = tileid.int_val();
      showcmd->SetTileId(tid);
    }
    dbcomp->Interpret(showcmd);

    VpfLibComp* libcomp = (VpfLibComp*)dbcomp->FindLibComp(lib);
    VpfCoverage* cov = lib->coverage(covname.string_ptr());
    VpfCovComp* covcomp = (VpfCovComp*)libcomp->FindCovComp(cov);
    VpfFeatureClass* fclass = cov->feature_class(fcname.string_ptr());
    VpfFClassComp* fccomp = (VpfFClassComp*)covcomp->FindFClassComp(fclass);
    ComponentView* view = new ComponentView(fccomp);
    result.type(AttributeValue::ObjectType);
    result.obj_ref() = view;
    result.obj_type_ref() = FEATURE_CLASS;
    push_stack(result);
}

#endif /* !defined(VPF_FREE) */

/*****************************************************************************/

#if !defined(DLG_FREE)

Load100KDLGFCFunc::Load100KDLGFCFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void Load100KDLGFCFunc::execute() {
    ComValue dbpath = stack_arg(0);
    ComValue libname = stack_arg(1);
    ComValue covname = stack_arg(2);
    ComValue fcname = stack_arg(3);
    reset_stack();
    ComValue result;

    Creator::instance(new DlgCreator());
    const char* dbdir = dbpath.string_ptr();
    DlgDatabase* db = new DlgDatabase(dbdir);
    DlgLibrary* lib = db->library(libname.string_ptr());
    DlgDbComp* dbcomp = new DlgDbComp(db);
    MapShowCmd* showcmd = new MapShowCmd();
    showcmd->SetUseEditor(false);
    showcmd->SetLibName(libname.string_ptr());
    showcmd->SetCovName(covname.string_ptr());
    showcmd->SetFClassName(fcname.string_ptr());
    dbcomp->Interpret(showcmd);

    DlgLibComp* libcomp = (DlgLibComp*)dbcomp->FindLibComp(lib);
    DlgCoverage* cov = lib->coverage(covname.string_ptr());
    DlgCovComp* covcomp = (DlgCovComp*)libcomp->FindCovComp(cov);
    DlgFeatureClass* fclass = cov->feature_class(fcname.string_ptr());
    DlgFClassComp* fccomp = (DlgFClassComp*)covcomp->FindFClassComp(fclass);
    ComponentView* view = new ComponentView(fccomp);
    result.type(AttributeValue::ObjectType);
    result.obj_ref() = view;
    result.obj_type_ref() = FEATURE_CLASS;
    push_stack(result);
}

#endif /* !defined(DLG_FREE) */

/*****************************************************************************/

FeatureFunc::FeatureFunc(ComTerp* comterp) : ComFunc(comterp) {
  _id_symid = symbol_add("id");
}

/*****************************************************************************/

PointFeatureFunc::PointFeatureFunc(ComTerp* comterp) : FeatureFunc(comterp) {
}

void PointFeatureFunc::execute() {
  ComValue x1(stack_arg(0));
  ComValue y1(stack_arg(1));
  ComValue fid(stack_key(_id_symid));
  reset_stack();
  float fx1 = x1.float_val();
  float fy1 = y1.float_val();
  MapPoint* mappoint = new MapPoint();
  mappoint->insert_points(1, &fx1, &fy1);
  MapFeature* point = new MapFeature(-1, MapFeature::PointType);
  if (fid.type() != ComValue::UnknownType) {
    int id = fid.int_val();
    point->id(id);
  }
  point->point_primitive(mappoint);
  mappoint->feature(point);
  MapFeatureComp* fcomp = new MapFeatureComp(point);
  ComponentView* view = new ComponentView(fcomp);
  ComValue result;
  result.type(AttributeValue::ObjectType);
  result.obj_ref() = view;
  result.obj_type_ref() = POINT_FEATURE;
  push_stack(result);
}

/*****************************************************************************/

LineFeatureFunc::LineFeatureFunc(ComTerp* comterp) : FeatureFunc(comterp) {
}

void LineFeatureFunc::execute() {
  ComValue fid(stack_key(_id_symid));
  int pn;
  pn = nargs() / 2;
  float* px = new float[pn];
  float* py = new float[pn];
  int ip = 0;
  for (int i = 0; i < nargs(); i += 2)
    px[ip++] = stack_arg(i).float_val();
  ip = 0;
  for (int i = 1; i < nargs(); i += 2)
    py[ip++] = stack_arg(i).float_val();
  reset_stack();
  MapEdge* mapedge = new MapEdge();
  mapedge->insert_pointers(pn, px, py, nil, true);
  MapFeature* line = new MapFeature(-1, MapFeature::LineType);
  if (fid.type() != ComValue::UnknownType) {
    int id = fid.int_val();
    line->id(id);
  }
  line->edge_primitive(mapedge);
  mapedge->feature(line);
  MapFeatureComp* fcomp = new MapFeatureComp(line);
  ComponentView* view = new ComponentView(fcomp);

  ComValue result;
  result.type(AttributeValue::ObjectType);
  result.obj_ref() = view;
  result.obj_type_ref() = LINE_FEATURE;
  push_stack(result);
}

/*****************************************************************************/

AreaFeatureFunc::AreaFeatureFunc(ComTerp* comterp) : FeatureFunc(comterp) {
}

void AreaFeatureFunc::execute() {
  ComValue fid(stack_key(_id_symid));
  int pn;
  pn = nargs() / 2;
  float* px = new float[pn];
  float* py = new float[pn];
  int ip = 0;
  for (int i = 0; i < nargs(); i += 2)
    px[ip++] = stack_arg(i).float_val();
  ip = 0;
  for (int i = 1; i < nargs(); i += 2)
    py[ip++] = stack_arg(i).float_val();
  MapFace* mapface = new MapFace();
  mapface->insert_pointers(pn, px, py, nil, true);
  MapFeature* area = new MapFeature(-1, MapFeature::AreaType);
  if (fid.type() != ComValue::UnknownType) {
    int id = fid.int_val();
    area->id(id);
  }
  area->face_primitive(mapface);
  mapface->feature(area);
  MapFeatureComp* fcomp = new MapFeatureComp(area);
  ComponentView* view = new ComponentView(fcomp);

  ComValue result;
  result.type(AttributeValue::ObjectType);
  result.obj_ref() = view;
  result.obj_type_ref() = AREA_FEATURE;
  push_stack(result);
}

/*****************************************************************************/

NodeFeatureFunc::NodeFeatureFunc(ComTerp* comterp) : FeatureFunc(comterp) {
}

void NodeFeatureFunc::execute() {
  ComValue fid(stack_key(_id_symid));
  ComValue x1(stack_arg(0));
  ComValue y1(stack_arg(1));
  reset_stack();
  float fx1 = x1.float_val();
  float fy1 = y1.float_val();
  MapNode* mapnode = new MapNode();
  mapnode->insert_points(1, &fx1, &fy1);
  MapFeature* node = new MapFeature(-1, MapFeature::JunctionType);
  if (fid.type() != ComValue::UnknownType) {
    int id = fid.int_val();
    node->id(id);
  }
  node->node_primitive(mapnode);
  mapnode->feature(node);
  MapFeatureComp* fcomp = new MapFeatureComp(node);
  ComponentView* view = new ComponentView(fcomp);

  ComValue result;
  result.type(AttributeValue::ObjectType);
  result.obj_ref() = view;
  result.obj_type_ref() = NODE_FEATURE;
  push_stack(result);
}

/*****************************************************************************/

PolygonAreaFunc::PolygonAreaFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void PolygonAreaFunc::execute() {
  ComValue mf(stack_arg(0));
  reset_stack();
  if (mf.type() == ComValue::ObjectType && mf.obj_type_val() == AREA_FEATURE) {
    ComponentView* view = (ComponentView*)mf.obj_val();
    MapFeatureComp* fcomp = (MapFeatureComp*)view->GetSubject();
    MapFeature* feat = fcomp->feature();
    MapFace* face = feat->face_primitive();
    float area = face->topo()->area();
    ComValue result(area);
    push_stack(result);
  }
  else {
    push_stack(ComValue::nullval());
  }
}

/*****************************************************************************/

FeatureClassFunc::FeatureClassFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void FeatureClassFunc::execute() {
  MapFeatureClass* fc = new MapFeatureClass(0);
  MapFClassComp* fcomp = new MapFClassComp(fc);
  ComponentView* view = new ComponentView(fcomp);
  ComValue result;
  result.type(AttributeValue::ObjectType);
  result.obj_ref() = view;
  result.obj_type_ref() = FEATURE_CLASS;
  push_stack(result);
}

/*****************************************************************************/

AddFeatureFunc::AddFeatureFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void AddFeatureFunc::execute() {
  ComValue feature(stack_arg(0));
  ComValue fclass(stack_arg(1));
  reset_stack();

  ComponentView* fview = (ComponentView*)feature.obj_val();
  ComponentView* fcview = (ComponentView*)fclass.obj_val();
  MapFeatureComp* fcomp = (MapFeatureComp*)fview->GetSubject();
  MapFClassComp* fccomp = (MapFClassComp*)fcview->GetSubject();
  MapFeature* f = fcomp->feature();
  MapFeatureClass* fc = fccomp->GetFeatureClass();
  fc->append(f);
  f->feature_class(fc);
  fccomp->Append(fcomp);

  ComValue result;
  result.type(AttributeValue::ObjectType);
  result.obj_ref() = fcview;
  result.obj_type_ref() = FEATURE_CLASS;
  push_stack(result);
}

/*****************************************************************************/

PointDistanceFunc::PointDistanceFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void PointDistanceFunc::execute() {
  ComValue pt1(stack_arg(0));
  ComValue pt2(stack_arg(1));
  reset_stack();
  if (pt1.type() == ComValue::ObjectType &&
      pt1.obj_type_val() == POINT_FEATURE &&
      pt2.type() == ComValue::ObjectType &&
      pt2.obj_type_val() == POINT_FEATURE) {
    ComponentView* view1 = (ComponentView*)pt1.obj_val();
    MapFeatureComp* fcomp1 = (MapFeatureComp*)view1->GetSubject();
    MapFeature* feat1 = fcomp1->feature();
    MapPoint* point1 = feat1->point_primitive();
    float x1 = point1->topo()->xpoints()[0];
    float y1 = point1->topo()->ypoints()[0];
    ComponentView* view2 = (ComponentView*)pt2.obj_val();
    MapFeatureComp* fcomp2 = (MapFeatureComp*)view2->GetSubject();
    MapFeature* feat2 = fcomp2->feature();
    MapPoint* point2 = feat2->point_primitive();
    float x2 = point2->topo()->xpoints()[0];
    float y2 = point2->topo()->ypoints()[0];

    float dist = sqrt((y2-y1)*(y2-y1)+(x2-x1)*(x2-x1));
    ComValue result(dist);
    push_stack(result);
  }
  else {
    push_stack(ComValue::nullval());
  }
}

/*****************************************************************************/

NearestPointInLineFunc::NearestPointInLineFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void NearestPointInLineFunc::execute () {
  ComValue pt(stack_arg(0));
  ComValue ln(stack_arg(1));
  reset_stack();

  if (pt.obj_type_val() == POINT_FEATURE && ln.obj_type_val() == LINE_FEATURE){
    ComponentView* ptview = (ComponentView*)pt.obj_val();
    ComponentView* lnview = (ComponentView*)ln.obj_val();
    MapFeatureComp* ptcomp = (MapFeatureComp*)ptview->GetSubject();
    MapFeatureComp* lncomp = (MapFeatureComp*)lnview->GetSubject();
    MapFeature* ptf = ptcomp->feature();
    MapFeature* lnf = lncomp->feature();
    MapPoint* ptp = ptf->point_primitive();
    MapEdge* lnp = lnf->edge_primitive();

    Point3 point;
    Line3 line;

    int nlpts = lnp->npts();
    double mindist = DBL_MAX;
    double mint = -1.0;
    int mini = 0;
    double it;
    const float* lx = lnp->xpoints();
    const float* ly = lnp->ypoints();
    for (int i = 1; i < nlpts; i++) {
      point.x = ptp->xpoints()[0] - lx[i-1];
      point.y = ptp->ypoints()[0] - ly[i-1];
      point.z = 0.0;
      line.b.x = 0.0;
      line.b.y = 0.0;
      line.b.z = 0.0;
      line.m.x = lx[i]-lx[i-1];
      line.m.y = ly[i]-ly[i-1];
      line.m.z = 0.0;
      double dist = MinPointLine(point, line, it);
      if (dist < mindist) {
	mindist = dist;
	mint = it;
	mini = i;
      }
    }
    if (mini > 0) {
      double dx = lx[mini]-lx[mini-1];
      double dy = ly[mini]-ly[mini-1];
      float rpx = dx * mint + lx[mini-1];
      float rpy = dy * mint + ly[mini-1];
      MapPoint* rpoint = new MapPoint();
      rpoint->insert_points(1, &rpx, &rpy);
      MapFeature* rfeat = new MapFeature(-1, MapFeature::PointType);
      rfeat->point_primitive(rpoint);
      rpoint->feature(rfeat);
      MapFeatureComp* rcomp = new MapFeatureComp(rfeat);
      ComponentView* rview = new ComponentView(rcomp);
      ComValue result;
      result.type(AttributeValue::ObjectType);
      result.obj_ref() = rview;
      result.obj_type_ref() = POINT_FEATURE;
      push_stack(result);
    }
    else
      push_stack(ComValue::nullval());
  }
  else
    push_stack(ComValue::nullval());
}

/*****************************************************************************/

ConvexHullFunc::ConvexHullFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void ConvexHullFunc::execute() {
  int nf = nargs();
  if (nf > 0) {
    int ptcount = 0;
    for (int i = 0; i < nf; i++) {
      ComValue val(stack_arg(i));
      if (val.type() == ComValue::ObjectType) {
	switch(val.obj_type_val()) {
	case POINT_FEATURE:
	  ptcount++;
	  break;
	case LINE_FEATURE: {
	  ComponentView* lview = (ComponentView*)val.obj_val();
	  MapFeatureComp* lcomp = (MapFeatureComp*)lview->GetSubject();
	  MapFeature* lf = lcomp->feature();
	  MapEdge* le = lf->edge_primitive();
	  ptcount += le->topo()->npts();
	}
	  break;
	case AREA_FEATURE: {
	  ComponentView* aview = (ComponentView*)val.obj_val();
	  MapFeatureComp* acomp = (MapFeatureComp*)aview->GetSubject();
	  MapFeature* af = acomp->feature();
	  MapFace* face = af->face_primitive();
	  ptcount += face->topo()->npts();
	}
	  break;
	case NODE_FEATURE:
	  ptcount++;
	  break;
	case FEATURE_CLASS: {
	  ComponentView* fcview = (ComponentView*)val.obj_val();
	  MapFClassComp* fccomp = (MapFClassComp*)fcview->GetSubject();
	  MapFeatureClass* fc = fccomp->GetFeatureClass();
	  for (int j = 0; j < fc->nfeats(); j++) {
	    MapFeature* f = fc->feature(j);
	    switch(f->type()) {
	    case MapFeature::PointType:
	      ptcount++;
	      break;
	    case MapFeature::LineType:
	      ptcount += f->edge_primitive()->topo()->npts();
	      break;
	    case MapFeature::AreaType:
	      ptcount += f->face_primitive()->topo()->npts();
	      break;
	    case MapFeature::JunctionType:
	      ptcount++;
	      break;
	    }
	  }
	}
	  break;
	}
      }
    }
    float* fx = new float[ptcount];
    float* fy = new float[ptcount];
    int pti = 0;
    for (int i = 0; i < nf; i++) {
      ComValue val(stack_arg(i));
      if (val.type() == ComValue::ObjectType) {
	switch(val.obj_type_val()) {
	case POINT_FEATURE: {
	  ComponentView* pview = (ComponentView*)val.obj_val();
	  MapFeatureComp* pcomp = (MapFeatureComp*)pview->GetSubject();
	  MapFeature* pf = pcomp->feature();
	  MapPoint* pp = pf->point_primitive();
	  fx[pti] = pp->topo()->xpoints()[0];
	  fy[pti] = pp->topo()->ypoints()[0];
	  pti++;
	}
	  break;
	case LINE_FEATURE: {
	  ComponentView* lview = (ComponentView*)val.obj_val();
	  MapFeatureComp* lcomp = (MapFeatureComp*)lview->GetSubject();
	  MapFeature* lf = lcomp->feature();
	  MapEdge* le = lf->edge_primitive();
	  for (int k = 0; k < le->topo()->npts(); k++) {
	    fx[pti] = le->topo()->xpoints()[k];
	    fy[pti] = le->topo()->ypoints()[k];
	    pti++;
	  }
	}
	  break;
	case AREA_FEATURE: {
	  ComponentView* aview = (ComponentView*)val.obj_val();
	  MapFeatureComp* acomp = (MapFeatureComp*)aview->GetSubject();
	  MapFeature* af = acomp->feature();
	  MapFace* face = af->face_primitive();
	  for (int k = 0; k < face->topo()->npts(); k++) {
	    fx[pti] = face->topo()->xpoints()[k];
	    fy[pti] = face->topo()->ypoints()[k];
	    pti++;
	  }
	}
	  break;
	case NODE_FEATURE: {
	  ComponentView* nview = (ComponentView*)val.obj_val();
	  MapFeatureComp* ncomp = (MapFeatureComp*)nview->GetSubject();
	  MapFeature* nf = ncomp->feature();
	  MapNode* np = nf->node_primitive();
	  fx[pti] = np->topo()->xpoints()[0];
	  fy[pti] = np->topo()->ypoints()[0];
	  pti++;
	}
	  break;
	case FEATURE_CLASS: {
	  ComponentView* fcview = (ComponentView*)val.obj_val();
	  MapFClassComp* fccomp = (MapFClassComp*)fcview->GetSubject();
	  MapFeatureClass* fc = fccomp->GetFeatureClass();
	  for (int j = 0; j < fc->nfeats(); j++) {
	    MapFeature* f = fc->feature(j);
	    switch(f->type()) {
	    case MapFeature::PointType:
	      fx[pti] = f->point_primitive()->topo()->xpoints()[0];
	      fy[pti] = f->point_primitive()->topo()->ypoints()[0];
	      pti++;
	      break;
	    case MapFeature::LineType:
	      for (int k = 0; k < f->edge_primitive()->topo()->npts(); k++) {
		fx[pti] = f->edge_primitive()->topo()->xpoints()[k];
		fy[pti] = f->edge_primitive()->topo()->ypoints()[k];
		pti++;
	      }
	      break;
	    case MapFeature::AreaType:
	      for (int k = 0; k < f->face_primitive()->topo()->npts(); k++) {
		fx[pti] = f->face_primitive()->topo()->xpoints()[k];
		fy[pti] = f->face_primitive()->topo()->ypoints()[k];
		pti++;
	      }
	      break;
	    case MapFeature::JunctionType:
	      fx[pti] = f->node_primitive()->topo()->xpoints()[0];
	      fy[pti] = f->node_primitive()->topo()->ypoints()[0];
	      pti++;
	      break;
	    }
	  }
	}
	  break;
	}
      }
    }
    reset_stack();
    float* hx;
    float* hy;
    int hn = ConvexHullCmd::ConvexHull(ptcount, fx, fy, hx, hy);
    delete [] fx;
    delete [] fy;
    if (hn > 0) {
      MapFace* hface = new MapFace();
      hface->insert_pointers(hn, hx, hy, nil, true);
      MapFeature* hfeat = new MapFeature(-1, MapFeature::AreaType);
      hfeat->face_primitive(hface);
      hface->feature(hfeat);
      MapFeatureComp* hcomp = new MapFeatureComp(hfeat);
      hfeat->graphic((void*)hcomp);
      ComponentView* hview = new ComponentView(hcomp);
      ComValue result;
      result.type(ComValue::ObjectType);
      result.obj_type_ref() = AREA_FEATURE;
      result.obj_ref() = hview;
      push_stack(result);
    }
    else
      push_stack(ComValue::nullval());
  }
  else {
    push_stack(ComValue::nullval());
  }
}

/*****************************************************************************/

ShapeSelFunc::ShapeSelFunc(ComTerp* comterp) : MapFunc(comterp) {
}

boolean ShapeSelFunc::node_in_node(MapFeature* pf1, MapFeature* pf2) {
  FPointObj* p1 = pf1->node_primitive()->topo()->point_obj();
  FPointObj* p2 = pf2->node_primitive()->topo()->point_obj();
  return (p1->_x == p2->_x && p1->_y == p2->_y);
}

boolean ShapeSelFunc::node_in_point(MapFeature* pf1, MapFeature* pf2) {
  FPointObj* p1 = pf1->node_primitive()->topo()->point_obj();
  FPointObj* p2 = pf2->point_primitive()->topo()->point();
  return (p1->_x == p2->_x && p1->_y == p2->_y);
}

boolean ShapeSelFunc::node_in_line(MapFeature* pf, MapFeature* lf) {
  FPointObj* p = pf->node_primitive()->topo()->point_obj();
  FMultiLineObj* l = lf->edge_primitive()->topo()->multiline();
  return l->Contains(*p);
}

boolean ShapeSelFunc::node_in_area(MapFeature* pf, MapFeature* af) {
  FPointObj* p = pf->node_primitive()->topo()->point_obj();
  FFillPolygonObj* pl = af->face_primitive()->topo()->polygon();
  return pl->Contains(*p);
}

boolean ShapeSelFunc::point_in_point(MapFeature* pf1, MapFeature* pf2) {
  FPointObj* p1 = pf1->point_primitive()->topo()->point();
  FPointObj* p2 = pf2->point_primitive()->topo()->point();
  return (p1->_x == p2->_x && p1->_y == p2->_y);
}

boolean ShapeSelFunc::point_in_line(MapFeature* pf, MapFeature* lf) {
  FPointObj* p = pf->point_primitive()->topo()->point();
  FMultiLineObj* l = lf->edge_primitive()->topo()->multiline();
  return l->Contains(*p);
}

boolean ShapeSelFunc::point_in_area(MapFeature* pf, MapFeature* af) {
  FPointObj* p = pf->point_primitive()->topo()->point();
  FFillPolygonObj* pl = af->face_primitive()->topo()->polygon();
  return pl->Contains(*p);
}

boolean ShapeSelFunc::line_in_line(MapFeature* lf1, MapFeature* lf2) {
  FMultiLineObj* l1 = lf1->edge_primitive()->topo()->multiline();
  FMultiLineObj* l2 = lf2->edge_primitive()->topo()->multiline();
  if (*l1 == *l2)
    return true;
  int count2 = l2->count();
  float* x2 = l2->x();
  float* y2 = l2->y();
  for (int i2 = 0; i2 < count2-1; i2++) {
    FLineObj tl(x2[i2], y2[i2], x2[i2+1], y2[i2+1]);
    if (l1->Intersects(tl))
      return true;
  }
  for (int i2 = 0; i2 < count2; i2++) {
    FPointObj tp(x2[i2], y2[i2]);
    if (l1->Contains(tp))
      return true;
  }

  return false;
}

boolean ShapeSelFunc::line_in_area(MapFeature* lf, MapFeature* af) {
  FMultiLineObj* l = lf->edge_primitive()->topo()->multiline();
  FFillPolygonObj* pl = af->face_primitive()->topo()->polygon();
  int count = l->count();
  float* x = l->x();
  float* y = l->y();
  for (int i = 0; i < count; i++) {
    FPointObj pt(x[i], y[i]);
    if (pl->Contains(pt))
      return true;
  }
  for (int i = 0; i < count-1; i++) {
    FLineObj ln(x[i], y[i], x[i+1], y[i+1]);
    if (pl->Intersects(ln))
      return true;
  }
  return false;
}

boolean ShapeSelFunc::area_in_area(MapFeature* af1, MapFeature* af2) {
  FFillPolygonObj* pl1 = af1->face_primitive()->topo()->polygon();
  FFillPolygonObj* pl2 = af2->face_primitive()->topo()->polygon();
  int count1 = pl1->count();
  float* x1 = pl1->x();
  float* y1 = pl1->y();
  int count2 = pl2->count();
  float* x2 = pl2->x();
  float* y2 = pl2->y();
  for (int i = 0; i < count1; i++) {
    FPointObj pt(x1[i], y1[i]);
    if (pl2->Contains(pt))
      return true;
  }
  for (int i = 0; i < count2; i++) {
    FPointObj pt(x2[i], y2[i]);
    if (pl1->Contains(pt))
      return true;
  }
  for (int i2 = 0; i2 < count2-1; i2++) {
    FLineObj tl(x2[i2], y2[i2], x2[i2+1], y2[i2+1]);
    if (pl1->Intersects(tl))
      return true;
  }

  return false;
}

void ShapeSelFunc::execute() {
  ComValue mfc(stack_arg(0));
  ComValue mf(stack_arg(1));
  reset_stack();
  if (mfc.type() == ComValue::ObjectType &&
      mfc.obj_type_val() == FEATURE_CLASS &&
      mf.type() == ComValue::ObjectType &&
      mf.obj_type_val() == AREA_FEATURE) {
    ComponentView* fcview = (ComponentView*)mfc.obj_val();
    MapFClassComp* fccomp = (MapFClassComp*)fcview->GetSubject();
    MapFeatureClass* fc = fccomp->GetFeatureClass();
    ComponentView* fview = (ComponentView*)mf.obj_val();
    MapFeatureComp* fcomp = (MapFeatureComp*)fview->GetSubject();
    MapFeature* feat = fcomp->feature();

    MapFeature** feats = new MapFeature*[fc->nfeats()];
    int nfeats = 0;
    int size = fc->nfeats();

    int nf = fc->nfeats();
    for (int i=0; i < nf; i++) {
      MapFeature* imf = fc->feature(i);
      switch(imf->type()) {
      case MapFeature::PointType:

	switch(feat->type()) {
	case MapFeature::PointType:
	  if (point_in_point(imf, feat)) {
	    MapFeature* newf = copy_point(imf,
					  (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::LineType:
	  if (point_in_line(imf, feat)) {
	    MapFeature* newf = copy_point(imf,
					  (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::AreaType:
	  if (point_in_area(imf, feat)) {
	    MapFeature* newf = copy_point(imf,
					  (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::JunctionType:
	  if (node_in_point(feat, imf)) {
	    MapFeature* newf = copy_point(imf,
					  (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	default:
	  break;
	}

	break;
      case MapFeature::LineType:

	switch(feat->type()) {
	case MapFeature::PointType:
	  if (point_in_line(feat, imf)) {
	    MapFeature* newf = copy_line(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::LineType:
	  if (line_in_line(feat, imf)) {
	    MapFeature* newf = copy_line(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::AreaType:
	  if (line_in_area(imf, feat)) {
	    MapFeature* newf = copy_line(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::JunctionType:
	  if (node_in_line(feat, imf)) {
	    MapFeature* newf = copy_line(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	default:
	  break;
	}

	break;
      case MapFeature::AreaType:

	switch(feat->type()) {
	case MapFeature::PointType:
	  if (point_in_area(feat, imf)) {
	    MapFeature* newf = copy_area(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::LineType:
	  if (line_in_area(feat, imf)) {
	    MapFeature* newf = copy_area(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::AreaType:
	  if (area_in_area(feat, imf)) {
	    MapFeature* newf = copy_area(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::JunctionType:
	  if (node_in_area(feat, imf)) {
	    MapFeature* newf = copy_area(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	default:
	  break;
	}

	break;
      case MapFeature::JunctionType:

	switch(feat->type()) {
	case MapFeature::PointType:
	  if (node_in_point(imf, feat)) {
	    MapFeature* newf = copy_junction(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::LineType:
	  if (node_in_line(imf, feat)) {
	    MapFeature* newf = copy_junction(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::AreaType:
	  if (node_in_area(imf, feat)) {
	    MapFeature* newf = copy_junction(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	case MapFeature::JunctionType:
	  if (node_in_node(imf, feat)) {
	    MapFeature* newf = copy_junction(imf,
					 (MapFeatureComp*)imf->graphic());
	    insert_array(feats, size, nfeats, newf);
	  }
	  break;
	default:
	  break;
	}

	break;
      default:
	break;
      }
    }

    MapFeatureClass* nfc = new MapFeatureClass(nfeats, feats);
    // assign fclass field of each feature
    for (int i = 0; i < nfeats; i++)
      feats[i]->feature_class(nfc);

    MapFClassComp* nfccomp = new MapFClassComp(nfc);
    for (int i = 0; i < nfeats; i++)
      nfccomp->Append((MapFeatureComp*)feats[i]->graphic());
    double dx1, dy1, dx2, dy2;
    int zone;
    nfc->extent(dx1, dy1, dx2, dy2, zone);
    AttributeList* alist = new AttributeList();
    AttributeValue* zoneval = new AttributeValue(zone, AttributeValue::IntType);
    alist->add_attr("zone", zoneval);
    AttributeValue* sweval = new AttributeValue(dx1);
    alist->add_attr("sw_e", sweval);
    AttributeValue* swnval = new AttributeValue(dy1);
    alist->add_attr("sw_n", swnval);
    AttributeValue* neeval = new AttributeValue(dx2);
    alist->add_attr("ne_e", neeval);
    AttributeValue* nenval = new AttributeValue(dy2);
    alist->add_attr("ne_n", nenval);
    nfccomp->SetAttributeList(alist);

    ComponentView* nview = new ComponentView(nfccomp);

    ComValue result;
    result.type(AttributeValue::ObjectType);
    result.obj_ref() = nview;
    result.obj_type_ref() = FEATURE_CLASS;

    push_stack(result);
  }
  else {
    push_stack(ComValue::nullval());
  }
}

/*****************************************************************************/

#if defined(CLIPPOLY)

ClipRectFunc::ClipRectFunc(ComTerp* comterp) : MapFunc(comterp) {
}

void ClipRectFunc::execute() {
    ComValue fcc(stack_arg(0));
    ComValue x0(stack_arg(1));
    ComValue y0(stack_arg(2));
    ComValue x1(stack_arg(3));
    ComValue y1(stack_arg(4));

    reset_stack();
    ComValue result;

    FBoxObj rect(x0.double_val(), y0.double_val(), x1.double_val(), y1.double_val());
    ComponentView* view = (ComponentView*)fcc.obj_val();
    MapFClassComp* fccomp = (MapFClassComp*)view->GetSubject();
    if (!fccomp) {
      push_stack(ComValue::nullval());
      return;
    }
    Iterator it;
    MapFeatureClass* fc = fccomp->GetFeatureClass();
    MapFeature** feats = new MapFeature*[fc->nfeats()];
    int nfeats = 0;
    int size = fc->nfeats();
    fccomp->First(it);
    for (int i = 0; i < fc->nfeats(); i++, fccomp->Next(it)) {
        MapFeature* f = fc->feature(i);
	MapFeatureComp* fcomp = (MapFeatureComp*)fccomp->GetComp(it);
	// if clip result then insert new features in feats[nfeats++]
	switch(f->type()) {
	case MapFeature::PointType: {
	  MapPoint* point = f->point_primitive();
	  if (point) {
	  FPointObj pt(point->xpoints()[0],
		      point->ypoints()[0]);
	  if (rect.Contains(pt)) {
	    MapFeature* newf = new MapFeature(f->id(), MapFeature::PointType);
	    newf->attrlist(new AttributeList(f->attrlist()));
	    AttributeList* compal = fcomp->GetAttributeList();
	    ALIterator alit;
	    for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
	      newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
	    MapPoint* newp = new MapPoint(newf);
	    newp->insert_points(1, (float*)point->xpoints(), (float*)point->ypoints());
	    newf->point_primitive(newp);
	    newf->graphic((void*)new MapFeatureComp(newf));
	    insert_array(feats, size, nfeats, newf);
	  }
	  }
	}
	  break;
	case MapFeature::LineType: {
	  MapEdge* edge = f->edge_primitive();
	  int npts = edge->npts();
	  float* fx = (float*)edge->xpoints();
	  float* fy = (float*)edge->ypoints();
	  int* ix = new int[npts];
	  int* iy = new int[npts];
	  for (int k = 0; k < npts; k++) {
	    ix[k] = Math::round(fx[k]);
	    iy[k] = Math::round(fy[k]);
	  }
	  FMultiLineObj line(fx, fy, npts);
	  if (line.Within(rect)) { // just copy it
	    MapFeature* newf = new MapFeature(f->id(), MapFeature::LineType);
	    newf->attrlist(new AttributeList(f->attrlist()));
	    AttributeList* compal = fcomp->GetAttributeList();
	    ALIterator alit;
	    for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
	      newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
	    MapEdge* newe = new MapEdge(newf);
	    newe->insert_points(npts, (float*)fx, (float*)fy);
	    newf->edge_primitive(newe);
	    newf->graphic((void*)new MapFeatureComp(newf));
	    insert_array(feats, size, nfeats, newf);
	  }
	  else if (line.Intersects(rect)) { // clip it
	    // this one still to be converted to floats
	    int nlines;
	    int* ni;
	    int** nx;
	    int** ny;
	    clipmultiline(npts, ix, iy, 
			  Math::round(rect._left),
			  Math::round(rect._bottom),
			  Math::round(rect._right),
			  Math::round(rect._top),
			  nlines, ni, nx, ny);
	    for (int j = 0; j < nlines; j++) {
	      MapFeature* newf = new MapFeature(f->id(), MapFeature::LineType);
	      newf->attrlist(new AttributeList(f->attrlist()));
	    AttributeList* compal = fcomp->GetAttributeList();
	    ALIterator alit;
	    for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
	      newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
	      MapEdge* newe = new MapEdge(newf);
	      newe->insert_points(ni[j], nx[j], ny[j]);
	      newf->edge_primitive(newe);
	      newf->graphic((void*)new MapFeatureComp(newf));
	      insert_array(feats, size, nfeats, newf);
	    }
	  }
	  delete [] ix;
	  delete [] iy;
	}
	  break;
	case MapFeature::AreaType: {
	  MapFace* face = f->face_primitive();
	  int npts = face->npts();
	  float* fx = (float*)face->xpoints();
	  float* fy = (float*)face->ypoints();
	  FFillPolygonObj poly(fx, fy, npts);
	  FBoxObj bbox;
	  poly.GetBox(bbox);
	  if (bbox.Within(rect)) {
	    MapFeature* newf = new MapFeature(f->id(), MapFeature::AreaType);
	    newf->attrlist(new AttributeList(f->attrlist()));
	    AttributeList* compal = fcomp->GetAttributeList();
	    ALIterator alit;
	    for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
	      newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
	    MapFace* newface = new MapFace(newf);
	    newface->insert_points(npts, (float*)fx, (float*)fy);
	    newf->face_primitive(newface);
	    newf->graphic((void*)new MapFeatureComp(newf));
	    insert_array(feats, size, nfeats, newf);
	  }
	  else if (poly.Intersects(rect)) {
	    int npolys;
	    int* ni;
	    float** nx;
	    float** ny;
	    float rx[4] = { rect._left, rect._left, rect._right, rect._right };
	    float ry[4] = { rect._bottom, rect._top, rect._top, rect._bottom };
	    clippoly(A_AND_B, npts, fx, fy, 4, rx, ry,
			  npolys, ni, nx, ny);
	    for (int j = 0; j < npolys; j++) {
	      MapFeature* newf = new MapFeature(f->id(), MapFeature::AreaType);
	      newf->attrlist(new AttributeList(f->attrlist()));
	    AttributeList* compal = fcomp->GetAttributeList();
	    ALIterator alit;
	    for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
	      newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
	      MapFace* newface = new MapFace(newf);
	      newface->insert_points(ni[j], nx[j], ny[j]);
	      newf->face_primitive(newface);
	      newf->graphic((void*)new MapFeatureComp(newf));
	      insert_array(feats, size, nfeats, newf);
	    }
	  }
	}
	  break;
	case MapFeature::JunctionType: {
	  MapNode* node = f->node_primitive();
	  if (node) {
	  FPointObj pt(node->xpoints()[0],
		      node->ypoints()[0]);
	  if (rect.Contains(pt)) {
	    MapFeature* newf = new MapFeature(f->id(), MapFeature::JunctionType);
	    newf->attrlist(new AttributeList(f->attrlist()));
	    AttributeList* compal = fcomp->GetAttributeList();
	    ALIterator alit;
	    for (compal->First(alit); !compal->Done(alit); compal->Next(alit))
	      newf->attrlist()->Append(new Attribute(*compal->GetAttr(alit)));
	    MapNode* newp = new MapNode(newf);
	    newp->insert_points(1, (float*)node->xpoints(), (float*)node->ypoints());
	    newf->node_primitive(newp);
	    newf->graphic((void*)new MapFeatureComp(newf));
	    insert_array(feats, size, nfeats, newf);
	  }
	  }
	}
	  break;
	}
    }
    MapFeatureClass* nfc = new MapFeatureClass(nfeats, feats);
    // assign fclass field of each feature
    double ox1, oy1, ox2, oy2;
    int utmzone;
    fc->extent(ox1, oy1, ox2, oy2, utmzone);
    nfc->set_extent(rect._left, rect._bottom, rect._right, rect._top, utmzone);
    for (int i = 0; i < nfeats; i++)
      feats[i]->feature_class(nfc);

    MapFClassComp* nfccomp = new MapFClassComp(nfc);
    for (int i = 0; i < nfeats; i++)
      nfccomp->Append((MapFeatureComp*)feats[i]->graphic());
    double dx1, dy1, dx2, dy2;
    int zone;
    nfc->extent(dx1, dy1, dx2, dy2, zone);
    AttributeList* alist = new AttributeList();
    AttributeValue* zoneval = new AttributeValue(zone, AttributeValue::IntType);
    alist->add_attr("zone", zoneval);
    AttributeValue* sweval = new AttributeValue(dx1);
    alist->add_attr("sw_e", sweval);
    AttributeValue* swnval = new AttributeValue(dy1);
    alist->add_attr("sw_n", swnval);
    AttributeValue* neeval = new AttributeValue(dx2);
    alist->add_attr("ne_e", neeval);
    AttributeValue* nenval = new AttributeValue(dy2);
    alist->add_attr("ne_n", nenval);
    nfccomp->SetAttributeList(alist);

    ComponentView* nview = new ComponentView(nfccomp);

    result.type(AttributeValue::ObjectType);
    result.obj_ref() = nview;
    result.obj_type_ref() = FEATURE_CLASS;
    push_stack(result);
}

/*****************************************************************************/

ClipPolyAminBFunc::ClipPolyAminBFunc(ComTerp* comterp) : ClipPolyFunc(comterp)
{
    _op = A_MIN_B;
}

/*****************************************************************************/

ClipPolyBminAFunc::ClipPolyBminAFunc(ComTerp* comterp) : ClipPolyFunc(comterp)
{
    _op = B_MIN_A;
}

/*****************************************************************************/

ClipPolyAandBFunc::ClipPolyAandBFunc(ComTerp* comterp) : ClipPolyFunc(comterp)
{
    _op = A_AND_B;
}

/*****************************************************************************/

ClipPolyFunc::ClipPolyFunc(ComTerp* comterp) : MapFunc(comterp) {
}

void ClipPolyFunc::execute() {
    ComValue fcc(stack_arg(0));
    ComValue poly(stack_arg(1));
    reset_stack();
    ComValue result;

    ComponentView* pview = (ComponentView*)poly.obj_val();
    MapFeatureComp* pcomp = (MapFeatureComp*)pview->GetSubject();
    MapFeature* mf = pcomp->feature();
    FFillPolygonObj* polyobj = nil;
    if (mf->type() == MapFeature::AreaType)
      polyobj = mf->face_primitive()->topo()->polygon();
    ComponentView* fcview = (ComponentView*)fcc.obj_val();
    MapFClassComp* fccomp = (MapFClassComp*)fcview->GetSubject();
    MapFeatureClass* fc = fccomp->GetFeatureClass();
    MapFeature** feats = new MapFeature*[fc->nfeats()];
    int nfeats = 0;
    int size = fc->nfeats();
    if (polyobj)
    for (int i = 0; i < fc->nfeats(); i++) {
        MapFeature* f = fc->feature(i);
	// if clip result then insert new features in feats[nfeats++]
	switch(f->type()) {
	case MapFeature::PointType: {
	  MapPoint* point = f->point_primitive();
	  FPointObj pt(point->xpoints()[0], point->ypoints()[0]);
	  if ((polyobj->Contains(pt) && _op == A_AND_B) ||
	      (!polyobj->Contains(pt) && _op != A_AND_B)) {
	    MapFeature* newf = new MapFeature(f->id(), MapFeature::PointType);
	    newf->attrlist(new AttributeList(f->attrlist()));
	    MapPoint* newp = new MapPoint(newf);
	    newp->insert_points(1, (float*)point->xpoints(), (float*)point->ypoints());
	    newf->point_primitive(newp);
	    newf->graphic((void*)new MapFeatureComp(newf));
	    insert_array(feats, size, nfeats, newf);
	  }
	}
	  break;
	case MapFeature::LineType: {
	  MapEdge* edge = f->edge_primitive();
	  int npts = edge->npts();
	  float* fx = (float*)edge->xpoints();
	  float* fy = (float*)edge->ypoints();
	  int* ix = new int[npts];
	  int* iy = new int[npts];
	  for (int k = 0; k < npts; k++) {
	    ix[k] = Math::round(fx[k]);
	    iy[k] = Math::round(fy[k]);
	  }
	  int* px = new int[polyobj->count()];
	  int* py = new int[polyobj->count()];
	  for (int k = 0; k < polyobj->count(); k++) {
	    px[k] = Math::round(polyobj->_x[k]);
	    py[k] = Math::round(polyobj->_y[k]);
	  }
	  int in_nlines;
	  int* in_ni;
	  float** in_cx;
	  float** in_cy;
	  int out_nlines;
	  int* out_ni;
	  float** out_cx;
	  float** out_cy;
	  cliplinepoly(npts, fx, fy, polyobj->count(), polyobj->_x, polyobj->_y,
			  in_nlines, in_ni, in_cx, in_cy,
			  out_nlines, out_ni, out_cx, out_cy);
	  if (_op == A_AND_B) {
	    if (in_nlines > 0) {
	      for (int k = 0; k < in_nlines; k++) {
		MapFeature* newf = new MapFeature(f->id(), MapFeature::LineType);
		newf->attrlist(new AttributeList(f->attrlist()));
		MapEdge* newe = new MapEdge(newf);
		newe->insert_points(in_ni[k], in_cx[k], in_cy[k]);
		newf->edge_primitive(newe);
		newf->graphic((void*)new MapFeatureComp(newf));
		insert_array(feats, size, nfeats, newf);
	      }
	    }
	  }
	  else {
	    if (out_nlines > 0) {
	      for (int k = 0; k < out_nlines; k++) {
		MapFeature* newf = new MapFeature(f->id(), MapFeature::LineType);
		newf->attrlist(new AttributeList(f->attrlist()));
		MapEdge* newe = new MapEdge(newf);
		newe->insert_points(out_ni[k], out_cx[k], out_cy[k]);
		newf->edge_primitive(newe);
		newf->graphic((void*)new MapFeatureComp(newf));
		insert_array(feats, size, nfeats, newf);
	      }
	    }
	  }
	}
	  break;
	case MapFeature::AreaType: {
	  MapFace* face = f->face_primitive();
	  int npts = face->npts();
	  float* fx = (float*)face->xpoints();
	  float* fy = (float*)face->ypoints();
	  int npolys;
	  int* ni;
	  float** cx;
	  float** cy;
	  clippoly(_op, polyobj->count(), polyobj->_x, polyobj->_y,
		   npts, fx, fy,
		   npolys, ni, cx, cy);
	  if (npolys > 0) {
	    for (int k = 0; k < npolys; k++) {
	    MapFeature* newf = new MapFeature(f->id(), MapFeature::AreaType);
	    newf->attrlist(new AttributeList(f->attrlist()));
	    MapFace* newface = new MapFace(newf);
	    newface->insert_points(ni[k], cx[k], cy[k]);
	    newf->face_primitive(newface);
	    newf->graphic((void*)new MapFeatureComp(newf));
	    insert_array(feats, size, nfeats, newf);
	    }
	  }
	}
	  break;
	case MapFeature::JunctionType: {
	  MapNode* point = f->node_primitive();
	  FPointObj pt(point->xpoints()[0], point->ypoints()[0]);
	  if ((polyobj->Contains(pt) && _op == A_AND_B) ||
	      (!polyobj->Contains(pt) && _op != A_AND_B)) {
	    MapFeature* newf = new MapFeature(f->id(), MapFeature::JunctionType);
	    newf->attrlist(new AttributeList(f->attrlist()));
	    MapNode* newp = new MapNode(newf);
	    newp->insert_points(1, (float*)point->xpoints(), (float*)point->ypoints());
	    newf->node_primitive(newp);
	    newf->graphic((void*)new MapFeatureComp(newf));
	    insert_array(feats, size, nfeats, newf);
	  }
	}
	  break;
	}
    }
    MapFeatureClass* nfc = new MapFeatureClass(nfeats, feats);
    // assign fclass field of each feature
    for (int i = 0; i < nfeats; i++)
      feats[i]->feature_class(nfc);

    MapFClassComp* nfccomp = new MapFClassComp(nfc);
    for (int i = 0; i < nfeats; i++)
      nfccomp->Append((MapFeatureComp*)feats[i]->graphic());
    ComponentView* nfcview = new ComponentView(nfccomp);

    result.type(AttributeValue::ObjectType);
    result.obj_ref() = nfcview;
    result.obj_type_ref() = FEATURE_CLASS;
    push_stack(result);
}

#endif /* defined(CLIPPOLY) */

/*****************************************************************************/

FclassExtentFunc::FclassExtentFunc(ComTerp* comterp) : ComFunc(comterp) {
    _swe_symid = symbol_add("swe");
    _swn_symid = symbol_add("swn");
    _nee_symid = symbol_add("nee");
    _nen_symid = symbol_add("nen");
}

void FclassExtentFunc::execute() {
    ComValue fcc(stack_arg(0));
    ComValue swe(stack_key(_swe_symid));
    ComValue swn(stack_key(_swn_symid));
    ComValue nee(stack_key(_nee_symid));
    ComValue nen(stack_key(_nen_symid));
    reset_stack();

    ComponentView* view = (ComponentView*)fcc.obj_val();
    MapFClassComp* comp = (MapFClassComp*)view->GetSubject();
    MapFeatureClass* fc = comp->GetFeatureClass();
    double x1, y1, x2, y2;
    int zone;
    fc->extent(x1, y1, x2, y2, zone);
    if (zone==0) {
      char lon0[80];
      sprintf(lon0, "lon_0=%f", x1);  // let proj pick the utm zone
      char* parms[4];
      parms[0] = "proj=utm";
      parms[1] = "ellps=WGS84";
      parms[2] = lon0;
      parms[3] = "no_defs";
      PJ* proj = pj_init(4, parms);
      UV data;
      data.u = x1 * DEG_TO_RAD;
      data.v = y1 * DEG_TO_RAD;
      data = pj_fwd(data, proj);
      x1 = data.u;
      y1 = data.v;
      data.u = x2 * DEG_TO_RAD;
      data.v = y2 * DEG_TO_RAD;
      data = pj_fwd(data, proj);
      x2 = data.u;
      y2 = data.v;
    } 
#if 0
    cout << "(" << Math::round(x1) << ", " << Math::round(y1) 
	 << ", " << Math::round(x2) << ", " << Math::round(y2) << ")\n";
#endif
    int nflags = swn.is_true() + swe.is_true() + nen.is_true() + nee.is_true();
    if (nflags==0 || nflags>1) {
      AttributeValueList* al = new AttributeValueList();
      if (swe.is_true() || !nflags) al->Append(new AttributeValue(x1));
      if (swn.is_true() || !nflags) al->Append(new AttributeValue(y1));
      if (nee.is_true() || !nflags) al->Append(new AttributeValue(x2));
      if (nen.is_true() || !nflags) al->Append(new AttributeValue(y2));
      ComValue retval(al);
      push_stack(retval); 
    } else {
      double result;
      if (swe.is_true()) result = x1;
      if (swn.is_true()) result = y1;
      if (nee.is_true()) result = x2;
      if (nen.is_true()) result = y2;
      ComValue retval(result);
      push_stack(retval);
    }
}

/*****************************************************************************/

AttrSelFunc::AttrSelFunc(ComTerp* comterp) : MapFunc(comterp) {
}


void AttrSelFunc::execute() {
    ComValue fcc(stack_arg(0));
    ComValue expr(stack_arg(1));
    reset_stack();
    ComValue result;

    ComTerpServ* attrterp = new ComTerpServ();
    attrterp->add_defaults();
    attrterp->read_string(symbol_pntr(expr.string_val()));

    ComponentView* view = (ComponentView*)fcc.obj_val();
    MapFClassComp* fccomp = (MapFClassComp*)view->GetSubject();
    MapFeatureClass* fc = fccomp->GetFeatureClass();
    MapFeature** feats = new MapFeature*[fc->nfeats()];
    int nfeats = 0;
    int size = fc->nfeats();
    Iterator i;
    for (fccomp->First(i); !fccomp->Done(i); fccomp->Next(i)) {
        MapFeatureComp* fcomp = (MapFeatureComp*)fccomp->GetComp(i);
	attrterp->set_attributes(fcomp->GetAttributeList());
	ComValue& value = attrterp->run(nil);
    	if (value.is_true()) {
	    MapFeature* newf = new MapFeature(*fcomp->feature());
	    newf->graphic((void*)new MapFeatureComp(newf));
            insert_array(feats, size, nfeats, newf);
	}
    }

    MapFeatureClass* nfc = new MapFeatureClass(nfeats, feats);
    // assign fclass field of each feature
    for (int j = 0; j < nfeats; j++)
      feats[j]->feature_class(nfc);

    MapFClassComp* nfccomp = new MapFClassComp(nfc);
    for (int j = 0; j < nfeats; j++)
      nfccomp->Append((MapFeatureComp*)feats[j]->graphic());
    ComponentView* nview = new ComponentView(nfccomp);

    result.type(AttributeValue::ObjectType);
    result.obj_ref() = nview;
    result.obj_type_ref() = FEATURE_CLASS;

    push_stack(result);
    delete attrterp;
    return;
}

/*****************************************************************************/

LineEvalFunc::LineEvalFunc(ComTerp* comterp) : MapFunc(comterp) {
}

void LineEvalFunc::execute() {
    ComValue fcc(stack_arg(0));
    ComValue expr(stack_arg(1));
    ComValue x1(stack_arg(2));
    ComValue y1(stack_arg(3));
    ComValue x2(stack_arg(4));
    ComValue y2(stack_arg(5));
    reset_stack();
    ComValue result;

    float fx1, fy1, fx2, fy2;
    fx1 = x1.float_val();
    fy1 = y1.float_val();
    fx2 = x2.float_val();
    fy2 = y2.float_val();
    FPointObj p1(fx1, fy1);
    FPointObj p2(fx2, fy2);
    FLineObj line(fx1, fy1, fx2, fy2);

    ComTerpServ* attrterp = new ComTerpServ();
    attrterp->add_defaults();
    attrterp->read_string(symbol_pntr(expr.string_val()));

    ComponentView* view = (ComponentView*)fcc.obj_val();
    MapFClassComp* fccomp = (MapFClassComp*)view->GetSubject();
    MapFeatureClass* fc = fccomp->GetFeatureClass();
    MapFeature** feats = new MapFeature*[fc->nfeats()];
    int nfeats = 0;
    int size = fc->nfeats();
    Iterator i;
    for (fccomp->First(i); !fccomp->Done(i); fccomp->Next(i)) {
        MapFeatureComp* fcomp = (MapFeatureComp*)fccomp->GetComp(i);
	MapFeature* feat = fcomp->feature();
	boolean touchline = false;
	switch (feat->type()) {
	case MapFeature::PointType: {
	  MapPoint* point = feat->point_primitive();
	  FPointObj fp(point->xpoints()[0], point->ypoints()[0]);
	  touchline = line.Same(fp, p1) || line.Same(fp, p2) ||
	    line.Contains(fp);
	}
	  break;
	case MapFeature::LineType: {
	  MapEdge* edge = feat->edge_primitive();
	  int npts = edge->npts();
	  float* fx = (float*)edge->xpoints();
	  float* fy = (float*)edge->ypoints();
	  FMultiLineObj fl(fx, fy, npts);
	  touchline = fl.Contains(p1) || fl.Contains(p2) ||
	    fl.Intersects(line);
	}
	  break;
	case MapFeature::AreaType: {
	  MapFace* face = feat->face_primitive();
	  int npts = face->npts();
	  float* fx = (float*)face->xpoints();
	  float* fy = (float*)face->ypoints();
	  FFillPolygonObj fp(fx, fy, npts);
	  touchline = fp.Contains(p1) || fp.Contains(p2) ||
	    fp.Intersects(line);
	}
	break;
	}

	if (touchline) {
	  attrterp->set_attributes(fcomp->GetAttributeList());
	  ComValue& value = attrterp->run(nil);
	  if (value.is_true()) {
	    MapFeature* newf = new MapFeature(*fcomp->feature());
	    newf->graphic((void*)new MapFeatureComp(newf));
            insert_array(feats, size, nfeats, newf);
	  }
	}
    }

    MapFeatureClass* nfc = new MapFeatureClass(nfeats, feats);
    // assign fclass field of each feature
    for (int j = 0; j < nfeats; j++)
      feats[j]->feature_class(nfc);

    MapFClassComp* nfccomp = new MapFClassComp(nfc);
    for (int j = 0; j < nfeats; j++)
      nfccomp->Append((MapFeatureComp*)feats[j]->graphic());
    ComponentView* nview = new ComponentView(nfccomp);

    result.type(AttributeValue::ObjectType);
    result.obj_ref() = nview;
    result.obj_type_ref() = FEATURE_CLASS;

    push_stack(result);
    delete attrterp;
    return;
}

/*****************************************************************************/

ExportFCFunc::ExportFCFunc(ComTerp* comterp) : ComFunc(comterp)
{
    _paint_symid = symbol_add("paint");
    _file_symid = symbol_add("file");
    _tileid_symid = symbol_add("tileid");
    _sock_symid = symbol_add("socket");
    _bitmap_symid = symbol_add("bitmap");
    _string_symid = symbol_add("string");
}

void ExportFCFunc::execute() {
    ComValue fclass(stack_arg(0));
    ComValue host(stack_arg(1));	
    ComValue port(stack_arg(2));	
    ComValue paint(stack_key(_paint_symid));
    ComValue file(stack_key(_file_symid));
    ComValue tileidval(stack_key(_tileid_symid));
    ComValue sock(stack_key(_sock_symid));
    ComValue bitmap(stack_key(_bitmap_symid));
    ComValue string(stack_key(_string_symid));
    reset_stack();
    if (nargs()==0 || fclass.null() || fclass.type() == ComValue::BlankType) {
        push_stack(ComValue::nullval());
        return;
    }

    if (paint.is_true()) {
        PaintFCFunc* pf = new PaintFCFunc(comterp());
        push_stack(fclass);
        pf->argcnts(1,0);
        pf->execute();
        delete pf;
    }

    int tileid;
    if (tileidval.is_type(ComValue::UnknownType))
      tileid = 1;
    else
      tileid = tileidval.int_val();

#ifdef HAVE_ACE
    ACE_SOCK_Stream* socket = nil;
#endif

    filebuf fbuf;
    if (file.is_type(ComValue::StringType))
        fbuf.open(file.string_ptr(), "w");

    else if (sock.is_true()) {
#ifdef HAVE_ACE
	ComTerpServ* terp = (ComTerpServ*)comterp();
	ComterpHandler* handler = (ComterpHandler*)terp->handler();
	if (handler) {
	  ACE_SOCK_Stream peer = handler->peer();
	  fbuf.attach(peer.get_handle());
	}
	else
#endif
	  fbuf.attach(fileno(stdout));
    }

    else {
#ifdef HAVE_ACE
        const char* hoststr = nil;
        const char* portstr = nil;
        hoststr = host.type()==ComValue::StringType ? host.string_ptr() : nil;
        portstr = port.type()==ComValue::StringType ? port.string_ptr() : nil;
        u_short portnum = portstr ? atoi(portstr) : port.ushort_val();
    
        if (portnum) {
            socket = new ACE_SOCK_Stream;
            ACE_SOCK_Connector conn;
            ACE_INET_Addr addr (portnum, hoststr);
    
            if (conn.connect (*socket, addr) == -1)
                ACE_ERROR ((LM_ERROR, "%p\n", "open"));
            fbuf.attach(socket->get_handle());
        } else if (comterp()->handler()) {
            fbuf.attach(comterp()->handler()->get_handle());
        } else
#endif
            fbuf.attach(fileno(stdout));
    }

    ComponentView* view = (ComponentView*)fclass.obj_val();
    MapFClassComp* fccomp = (MapFClassComp*)view->GetSubject();
    if (bitmap.is_true()) {
      int ncols, nrows;
      float scale = 1.;
      float xorigin, yorigin;
      int utmzone;
      unsigned char* bitmap = fccomp->bitmap(scale, ncols, nrows,
					     xorigin, yorigin, utmzone);
      if (bitmap) {
	ostream out(&fbuf);
#define PGM_OUT 0
#if !PGM_OUT
	out << "P4\n";
#else
	out << "P5\n";
#endif
	out << "# CREATOR: ivmaps mapserv\n";
	out << "# scale " << scale << "\n";
	out << "# xorigin " << Math::round(xorigin) << "\n";
	out << "# yorigin " << Math::round(yorigin) << "\n";
	out << "# utmzone " << utmzone << "\n";
#if !PGM_OUT
	out << ncols << "\n";
	out << nrows << "\n";
#else
	out << ncols << " " << nrows << "\n";
	out << 255 << "\n";
#endif
#if !PGM_OUT
	for (int row=0; row<nrows; row++) {
	  for (int col=0; col<ncols; col+=8) {
	    unsigned char byte = 0x00;
	    unsigned char mask = 0x80;
	    for (int bit=0; bit<8; bit++) {
	      if (bit+col<ncols && bitmap[bit+col+row*ncols]) 
		byte = byte | mask;
	      mask = mask >> 1;
	    }
	    out << byte;
	  }
	}
#else
	for (int row=0; row<nrows; row++) {
	  for (int col=0; col<ncols; col++) {
	    unsigned char byte = bitmap[col+row*ncols] ? 0xff : 0x00;
	    out << byte;
	  }
	}
#endif
	out.flush();
	delete bitmap;
      }
    } else {
    
      OverlayPS* ovpsv = (OverlayPS*) fccomp->Create(RAWMAP_VIEW);
      fccomp->Attach(ovpsv);
#if !defined(VPF_FREE)
      if (ovpsv->IsA(VPFFCLASS_RAW))
	((VpfFClassRawScript*)ovpsv)->TileId(tileid);
#endif
      ovpsv->Update();
      ostream* out;
      if (string.is_true())
	out = new strstream();
      else
	out = new ostream(&fbuf);
      *out << "mapfcl(\n";
      for (int i=0; i<PaintFCTblFunc::_table_len; i++) 
        *out << "    gs( " << symbol_pntr(PaintFCTblFunc::_gs_table[i]) << "),\n";
      ovpsv->Definition(*out);
      AttributeList* attrlist = fccomp->GetAttributeList();
      *out << *attrlist;
      *out << ")\n";
      out->flush();

      if (string.is_true()) {
	char* str = ((strstream*)out)->str();
	istrstream* istr = new istrstream(str, strlen(str));
	ComValue result(ComValue::ObjectType, (void*)istr);
	push_stack(result);
      }
      delete out;
    }
    
#ifdef HAVE_ACE
    if (sock.is_false() && socket) {
      if (socket->close () == -1)
	ACE_ERROR ((LM_ERROR, "%p\n", "close"));
      delete(socket);
    }
#endif
}


/*****************************************************************************/

ExportFCGrFunc::ExportFCGrFunc(ComTerp* comterp) : ComFunc(comterp)
{
    _gs_str_id = symbol_add(":fillbg 1 :brush 65535,0 :fgcolor \"Black\",0,0,0 :bgcolor \"White\",1,1,1 :nonepat");
    _tr_str_id = symbol_add(":transform 1,0,0,1,0,0");
    _gs_symid = symbol_add("gs");
}

void ExportFCGrFunc::execute() {
    ComValue* fccp = nil;
    ComValue* hostp = nil;
    ComValue* portp = nil;
    if(nargs()>0) {
        fccp = new ComValue(stack_arg(0));
    } else {
        reset_stack();
        push_stack(ComValue::nullval());
        return;
    }

    if (nargs()>1) {
        hostp = new ComValue(stack_arg(1));
	if (nargs()>2) 
	    portp = new ComValue(stack_arg(2));
    }
    reset_stack();

    ComponentView* view = (ComponentView*)fccp->obj_val();
    MapFClassComp* fccomp = (MapFClassComp*)view->GetSubject();

    filebuf fbuf;
#ifdef HAVE_ACE
    const char* hoststr = nil;
    const char* portstr = nil;
    if (nargs()>=3) 
        hoststr = hostp->type()==ComValue::StringType ? hostp->string_ptr() : nil;
    if (nargs()>=2) 
        portstr = portp->type()==ComValue::StringType ? portp->string_ptr() : nil;
    u_short portnum = portstr ? atoi(portstr) : portp->ushort_val();

    ACE_SOCK_Stream* socket = nil;
    if (portnum) {
        socket = new ACE_SOCK_Stream;
        ACE_SOCK_Connector conn;
        ACE_INET_Addr addr (portnum, hoststr);

        if (conn.connect (*socket, addr) == -1)
            ACE_ERROR ((LM_ERROR, "%p\n", "open"));
        fbuf.attach(socket->get_handle());
    } else 
#endif
    
    fbuf.attach(fileno(stdout));

    ostream out(&fbuf);

    out << "drawtool(\n";
    out << "    gs(" << symbol_pntr(_gs_str_id) << "),\n";

    Iterator i;
    for (fccomp->First(i); !fccomp->Done(i);) {
        MapFeatureComp* fcomp = (MapFeatureComp*)fccomp->GetComp(i);
	AttributeList* alist = fcomp->GetAttributeList();
	Attribute* attr = nil;
	if (!alist->find(_gs_symid)) {
	  AttributeValue* val = new AttributeValue();
	  val->type(AttributeValue::IntType);
	  val->int_ref() = 0;
	  attr = new Attribute("gs", val);
	  alist->Append(attr);
	}
	OverlayScript* fview = (OverlayScript*)fcomp->Create(RAWMAP_VIEW);
	fcomp->Attach(fview);
	MapFeatureRawScript::native(false);
	fview->Update();
	out << "    ";
	fview->Definition(out);
	MapFeatureRawScript::native(true);
	fccomp->Next(i);
	if (!fccomp->Done(i)) 
	  out << ",\n";
	if (attr)
	  alist->Remove(attr);
	delete fview;
    }
    out << " " << symbol_pntr(_tr_str_id);
    out << "\n)\n";
    out.flush();

#ifdef HAVE_ACE
    if (socket && socket->close () == -1)
            ACE_ERROR ((LM_ERROR, "%p\n", "close"));
    delete(socket);
#endif

    push_stack(*fccp);
    delete fccp;
    delete hostp;
    delete portp;
}

/*****************************************************************************/


ExportFCGrGsFunc::ExportFCGrGsFunc
(ExportFCGrFunc* func) : ComFunc(func->comterp())
{
    _func = func;
}

void ExportFCGrGsFunc::execute() {
    if (nargs()) {
        ComValue& gs_arg = stack_arg(0);
        _func->_gs_str_id = gs_arg.string_val();
        reset_stack();
        push_stack(gs_arg);
    } else {
        void* value = nil;
	comterp()->localtable()->find(value, _func->_gs_str_id);
        reset_stack();
        push_stack(value ? *(ComValue*)value : ComValue::nullval()) ;
    }
}

/*****************************************************************************/

PrintFCFunc::PrintFCFunc(ComTerp* comterp) : ExportFCFunc(comterp)
{
}

void PrintFCFunc::execute() {
    ExportFCFunc::execute();
}

/*****************************************************************************/

ExportFCGrTrFunc::ExportFCGrTrFunc
(ExportFCGrFunc* func) : ComFunc(func->comterp())
{
    _func = func;
}

void ExportFCGrTrFunc::execute() {
    if (nargs()) {
        ComValue& tr_arg = stack_arg(0);
        _func->_tr_str_id = tr_arg.string_val();
        reset_stack();
        push_stack(tr_arg);
    } else {
        void* value = nil;
	comterp()->localtable()->find(value, _func->_tr_str_id);
        reset_stack();
        push_stack(value ? *(ComValue*)value : ComValue::nullval()) ;
    }
}

/*****************************************************************************/

PrintFCGrFunc::PrintFCGrFunc(ComTerp* comterp) : ExportFCGrFunc(comterp)
{
}


void PrintFCGrFunc::execute() {
    if (nargs()>=2)
        for (int i=0; i<nargs()-1; i++)
            pop_stack();
    argcnts(1,nkeys());
    ExportFCGrFunc::execute();
}

#if 0
/*****************************************************************************/

PrintFCBackFunc::PrintFCBackFunc(ComTerp* comterp) : ComFunc(comterp)
{
}


void PrintFCBackFunc::execute() {
    ComValue fcc(stack_arg(0));
    reset_stack();

    ComTerpServ* terp = (ComTerpServ*)comterp();
    ComterpHandler* handler = (ComterpHandler*)terp->handler();
    if (handler) {
        MapFClassComp* fccomp = (MapFClassComp*)fcc.obj_val();
	OverlayPS* ovpsv = (OverlayPS*) fccomp->Create(RAWMAP_VIEW);
	fccomp->Attach(ovpsv);
	ovpsv->Update();
	filebuf obuf(handler->get_handle());
	ostream ostr(&obuf);
	ostr << "mapfcl(\n";
	ovpsv->Definition(ostr);
	ostr << ")\n";
	ostr.flush();
    }
}
#endif

/*****************************************************************************/

PaintFCFunc::PaintFCFunc(ComTerp* comterp) : ComFunc(comterp)
{
    _gs_symid = symbol_add("gs");
}

void PaintFCFunc::execute() {
    ComValue fclass(stack_arg(0));
    reset_stack();
    if (nargs()==0 || fclass.null()) {
        push_stack(ComValue::nullval());
        return;
    }

    ComponentView* view = (ComponentView*)fclass.obj_val();
    MapFClassComp* fccomp = (MapFClassComp*)view->GetSubject();

    Iterator i;
    for (fccomp->First(i); !fccomp->Done(i); fccomp->Next(i)) {
        MapFeatureComp* fcomp = (MapFeatureComp*)fccomp->GetComp(i);
	AttributeList* alist = fcomp->GetAttributeList();
	Attribute* attr = nil;
	if (!alist->find(_gs_symid)) {
	    int id = PaintFCTblFunc::eval_exprs(alist);
	    if (id>=0) {
	        AttributeValue* val = new AttributeValue();
	        val->type(AttributeValue::IntType);
	        val->int_ref() = id;
	        attr = new Attribute("gs", val);
	        alist->Append(attr);
	    }
	}
	
    }
}

/*****************************************************************************/

int PaintFCTblFunc::_table_siz = 0;
int PaintFCTblFunc::_table_len = 0;
int* PaintFCTblFunc::_gs_table = nil;
postfix_token** PaintFCTblFunc::_expr_table = nil;
int* PaintFCTblFunc::_exprlen_table = nil;
ComTerpServ* PaintFCTblFunc::_gsterp = nil;

PaintFCTblFunc::PaintFCTblFunc(ComTerp* comterp) : ComFunc(comterp) {
    if (_table_siz == 0) init_tables();
    _reset_symid = symbol_add("reset");
}

void PaintFCTblFunc::init_tables() {
    if (_table_siz == 0) {
        _table_siz = 64;
        _gs_table = new int[_table_siz];
        _expr_table = new postfix_token*[_table_siz];
        _exprlen_table = new int[_table_siz];
    }
    _gsterp = new ComTerpServ();
    _gsterp->add_defaults();
}

void PaintFCTblFunc::reset_tables() {
    _table_len = 0;
}
      
void PaintFCTblFunc::grow_tables(int gsid, postfix_token* tokbuf, int numtok) {
    if (_table_len==_table_siz) {
        _table_siz *= 2;
        int* newgs = new int[_table_siz];
        postfix_token** newexpr = new postfix_token*[_table_siz];
        int* newexprlen = new int[_table_siz];
        for (int i=0; i<_table_len; i++) {
	    newgs[i] = _gs_table[i];
	    newexpr[i] = _expr_table[i];
	    newexprlen[i] = _exprlen_table[i];
        }
	delete _gs_table;
	_gs_table = newgs;
	delete _expr_table;
	_expr_table = newexpr;
	delete _exprlen_table;
	_exprlen_table = newexprlen;
    }

    _gs_table[_table_len] = gsid;  
    _expr_table[_table_len] = tokbuf;  
    _exprlen_table[_table_len] = numtok;  
    _table_len++;
}

int PaintFCTblFunc::eval_exprs(AttributeList* alist) {
    boolean found = false;
    int curr = 0;
    while (!found && curr<_table_len) {
        _gsterp->set_attributes(alist);
	ComValue& result = 
	    _gsterp->run(_expr_table[curr], _exprlen_table[curr]);
	if (result.is_true())
	    return curr;
        curr++;
    }
    return -1;
}
     
void PaintFCTblFunc::execute() {
    if (nargs()>=2) {
        ComValue expr_arg(stack_arg(0));
        ComValue gs_arg(stack_arg(1));
	reset_stack();
	if (gs_arg.type() == ComValue::StringType && 
	    expr_arg.type() == ComValue::StringType) {

	    int numtok = 0;
	    postfix_token* tokbuf = 
	        _gsterp->gen_code(symbol_pntr(expr_arg.string_val()), numtok);
	    if (numtok>0) 
	        grow_tables(gs_arg.string_val(),
					tokbuf, numtok);
	}
    } else if (nargs()==0) {
        ComValue reset_flag(stack_key(_reset_symid));
	reset_stack();
	if (reset_flag.is_true())
	    reset_tables();
	else {
	    for (int i=0; i<_table_len; i++) {
                cout << symbol_pntr(_gs_table[i]) << "\n";
	        postfix_token* expr = _expr_table[i];
	        for (int j=0; j<_exprlen_table[i]; j++) {
		    cout << "    ";
		    print_pfbuf(expr, j);
		}
	    }
	    if (!_table_len) cout << "empty paint table";
	}
    }

}

/*****************************************************************************/

ExprFCFunc::ExprFCFunc(ComTerp* comterp) : ComFunc(comterp)
{
    _terp = new ComTerpServ();
    _terp->add_defaults();
}

void ExprFCFunc::execute() {
    ComValue fclass(stack_arg(0));
    ComValue attrname(stack_arg(1, true));
    ComValue expr(stack_arg(2));
    reset_stack();
    if (nargs()==0) {
        push_stack(ComValue::nullval());
        return;
    }

    ComponentView* view = (ComponentView*)fclass.obj_val();
    MapFClassComp* fccomp = (MapFClassComp*)view->GetSubject();
    int attrname_symid = attrname.symbol_val();

    _terp->read_string(expr.string_ptr());
    Iterator i;
    for (fccomp->First(i); !fccomp->Done(i); fccomp->Next(i)) {
        MapFeatureComp* fcomp = (MapFeatureComp*)fccomp->GetComp(i);
	AttributeList* alist = fcomp->GetAttributeList();
	_terp->set_attributes(alist);
	Attribute* attr = nil;
	AttributeValue* val = alist->find(attrname_symid);
	if (!val) {
	    val = new AttributeValue();
	    attr = new Attribute(attrname.string_ptr(), val);
	    alist->Append(attr);
	}
	*val = _terp->run(nil);
    }
}

/*****************************************************************************/

ImportFCFunc::ImportFCFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void ImportFCFunc::execute() {
    ComValue path(stack_arg(0));
    static int unidraw_symid = symbol_add("unidraw");
    ComValue use_unidraw(stack_key(unidraw_symid));
    reset_stack();

    Component* comp = nil;
    if (use_unidraw.is_false()) {
      MapCreator creator;
      MapCatalog catalog("import", &creator);
      boolean uflag = Component::use_unidraw();
      Component::use_unidraw(false);
      catalog.Retrieve(path.string_ptr(), comp);
      Component::use_unidraw(uflag);
    } else {
      unidraw->GetCatalog()->Retrieve(path.string_ptr(), comp);
    }

    if (comp && comp->IsA(MAPFCLASS_COMP)) {
      ComponentView* view = new ComponentView(comp);
      ComValue result(ComValue::ObjectType, (void*)view);
      result.obj_type_ref() = FEATURE_CLASS;
      push_stack(result);
    } else {
      if (comp) delete comp;
      push_stack(ComValue::nullval());
    }
}

/*****************************************************************************/

NumFeaturesFunc::NumFeaturesFunc(ComTerp* comterp) : ComFunc(comterp) {
}

void NumFeaturesFunc::execute() {
  ComValue fclass(stack_arg(0));
  reset_stack();
  if (fclass.type() == ComValue::ObjectType && 
      fclass.obj_type_val() == FEATURE_CLASS) {
    ComponentView* view = (ComponentView*)fclass.obj_val();
    MapFClassComp* comp = (MapFClassComp*)view->GetSubject();
    MapFeatureClass* fc = comp->GetFeatureClass();
    ComValue result(fc->nfeats());
    push_stack(result);
  }
  else
    push_stack(ComValue::nullval());
}

/*****************************************************************************/

GetFeatureFunc::GetFeatureFunc(ComTerp* comterp) : MapFunc(comterp) {
}

void GetFeatureFunc::execute() {
  ComValue fclass(stack_arg(0));
  ComValue fn(stack_arg(1));
  reset_stack();
  if (fclass.type() == ComValue::ObjectType && 
      fclass.obj_type_val() == FEATURE_CLASS &&
      fn.type() == ComValue::IntType) {
    ComponentView* view = (ComponentView*)fclass.obj_val();
    MapFClassComp* comp = (MapFClassComp*)view->GetSubject();
    MapFeatureClass* fc = comp->GetFeatureClass();
    int i = fn.int_val();
    MapFeature* feature = fc->feature(i);
    if (feature) {
      MapFeatureComp* fcomp = (MapFeatureComp*)feature->graphic();
      MapFeature* newf = nil;
      ComValue result;
      result.type(ComValue::ObjectType);
      switch(feature->type()) {
      case MapFeature::PointType:
	result.obj_type_ref() = POINT_FEATURE;
	newf = copy_point(feature, (MapFeatureComp*)feature->graphic());
	break;
      case MapFeature::LineType:
	result.obj_type_ref() = LINE_FEATURE;
	newf = copy_line(feature, (MapFeatureComp*)feature->graphic());
	break;
      case MapFeature::AreaType:
	result.obj_type_ref() = AREA_FEATURE;
	newf = copy_area(feature, (MapFeatureComp*)feature->graphic());
	break;
      case MapFeature::JunctionType:
	result.obj_type_ref() = NODE_FEATURE;
	newf = copy_junction(feature, (MapFeatureComp*)feature->graphic());
	break;
      }
      MapFeatureComp* nfcomp = (MapFeatureComp*)newf->graphic();
      ComponentView* fview = new ComponentView(nfcomp);
      result.obj_ref() = fview;
      push_stack(result);
    }
    else
      push_stack(ComValue::nullval());
  }
  else
    push_stack(ComValue::nullval());
}

/*****************************************************************************/

MapPasteFunc::MapPasteFunc(ComTerp* comterp, Editor* ed, OverlayCatalog* catalog) : PasteFunc(comterp, ed, catalog) {
  _sens_symid = symbol_add("sensitive");
}

void MapPasteFunc::SensitizeAll(OverlayComp* comp, boolean val) {
  if (comp && comp->IsA(OVERLAYS_COMP)) {
    Iterator i;
    for (comp->First(i); !comp->Done(i); comp->Next(i)) {
      OverlayComp* c = (OverlayComp*)comp->GetComp(i);
      SensitizeAll(c, val);
    }
    comp->GetAttributeList()->
      add_attr("sensitive",
	       new AttributeValue(val, AttributeValue::BooleanType));
  }
  else if (comp) {
    comp->GetAttributeList()->
      add_attr("sensitive",
	       new AttributeValue(val, AttributeValue::BooleanType));
  }
}

void MapPasteFunc::execute() {
    ComValue viewv(stack_arg(0));
    ComValue scalev(stack_arg(1));
    ComValue centerlon(stack_arg(2));
    ComValue centerlat(stack_arg(3));
    ComValue sensval(stack_key(_sens_symid));
    reset_stack();

    /* extract comp and scale/translate before pasting */
    ComponentView* view = (ComponentView*)viewv.obj_val();
    OverlayComp* comp = (OverlayComp*)view->GetSubject();

    if (scalev.is_num()) {
      Transformer* dt = ((MapEditor*)_ed)->GetMapViewState()->deftrans();
      dt->scale(scalev.float_val(), scalev.float_val());
      ((MapEditor*)_ed)->GetMapViewState()->settrans(dt);
    }

    if (sensval.is_known())
      SensitizeAll(comp, sensval.is_true());
    
    /* set creator for gvupdater to use and disable unidraw (!use_unidraw) */
    /* then restore server/viewer state (use_unidraw) and original creator */
    PasteCmd* cmd = new PasteCmd(_ed, new Clipboard(comp));
    Creator* oldcreator = Creator::instance();
    if (_catalog && _catalog->GetCreator()) 
      Creator::instance(_catalog->GetCreator());
    boolean uflag = Component::use_unidraw();	
    Component::use_unidraw(_catalog ? false : true);
    execute_log(cmd);
    Creator::instance(oldcreator);
    Component::use_unidraw(uflag);

    if (centerlon.is_num() && centerlat.is_num()) {
      MapCenterCmd* cencmd = new MapCenterCmd(_ed);
      cencmd->CenterLonLat(centerlon.float_val(), centerlat.float_val());
      execute_log(cencmd);
    }

    push_stack(viewv);
}
