/*
 * Copyright 1997, 1998 Vectaport Inc.
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <TopoFace/topoedge.h>
#include <TopoFace/topoedgelist.h>
#include <TopoFace/toponode.h>
#include <Map/mapelt.h>
#include <Map/mapprim.h>
#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapclasses.h>
#include <MapUnidraw/mapeditor.h>
#include <MapUnidraw/mapinspect.h>
#include <MapUnidraw/mapstatecmds.h>
#include <MapUnidraw/mapstates.h>
#include <MapUnidraw/mapviews.h>

#include <Unidraw/editor.h>
#include <Unidraw/globals.h>
#include <Unidraw/iterator.h>
#include <Unidraw/manips.h>
#include <Unidraw/selection.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/viewer.h>
#include <Unidraw/Components/grcomp.h>
#include <Unidraw/Components/grview.h>

#include <IVGlyph/observables.h>

#include <InterViews/cursor.h>
#include <InterViews/event.h>
#include <InterViews/window.h>
#include <IV-2_6/InterViews/rubrect.h>

#include <IV-2_6/_enter.h>

#include <Attribute/attrlist.h>

#include <stdio.h>
#include <string.h>
#include <strstream.h>

/*****************************************************************************/

ClassId MapInspectTool::GetClassId () { return SELECT_TOOL; }

boolean MapInspectTool::IsA (ClassId id) {
    return SELECT_TOOL == id || Tool::IsA(id);
}

MapInspectTool::MapInspectTool (MapSelectionState* ss, ControlInfo* m) : Tool(m)
{
    selstate = ss;
}

Tool* MapInspectTool::Copy () { return new MapInspectTool(selstate,CopyControlInfo()); }

Manipulator* MapInspectTool::CreateManipulator (
    Viewer* v, Event& e, Transformer* rel
) {
    Manipulator* m = nil;
    OverlaysView* views = (OverlaysView*)v->GetGraphicView();
    Selection* s = v->GetSelection(), *newSel;
    Command* cmd;

    if (e.shift) {
        Localize(s, v);
    } else {
	s->Clear();
    }
    DoInspect(FindFeature(views, e, s), v);
    return m;
}

Command* MapInspectTool::InterpretManipulator (Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    Viewer* viewer = dm->GetViewer();
    GraphicView* views = viewer->GetGraphicView();
    Selection* s = viewer->GetSelection();
    RubberRect* rr = (RubberRect*) dm->GetRubberband();
    Selection* newSel;
    Coord l, b, r, t;
    Command* ret = nil;

    return ret;
}

void MapInspectTool::Localize (Selection* s, Viewer* v) {
    Iterator i;

    for (s->First(i); !s->Done(i);) {
        GraphicView* view = s->GetView(i);

        if (view->GetViewer() != v) {
            s->Remove(i);
            view->EraseHandles();

        } else {
            s->Next(i);
        }
    }
}

void MapInspectTool::DoInspect(GraphicView* view, Viewer* v) {
  if (view == nil) {
    if (selstate)
        selstate->mapselection("");
    return;
  }
  GraphicComp* grcomp = view->GetGraphicComp();
  if (grcomp->IsA(MAPFEATURE_COMP)) {
    MapFeatureComp* mfc = (MapFeatureComp*)grcomp;
    MapFeature* mf = mfc->feature();
    const char* desc = mf->describe();
    
    char buf[BUFSIZ];
    buf[0] = '\0';
    ostrstream ostr(buf, BUFSIZ);
    ostr << *mfc->GetAttributeList();
    ostr << '\0';
    char* inspstr = new char[strlen(buf)+strlen(desc)+1];
    strcpy(inspstr, desc);
    strcat(inspstr, buf);

    if (selstate) 
      selstate->mapselection(inspstr);
    else {
      ((OverlayEditor*)v->GetEditor())->MouseDocObservable()->textvalue("");
      ((OverlayEditor*)v->GetEditor())->AttrEdit(((OverlayView*)view)->GetOverlayComp());
      ((OverlayEditor*)v->GetEditor())->MouseDocObservable()->textvalue(OverlayKit::mouse_attr);
    }
  }
}

GraphicView* MapInspectTool::FindInList(OverlaysView* vs, Event& e) {
  GraphicView* view = nil;
  Iterator i;
  for (vs->Last(i); !vs->Done(i); vs->Prev(i)) {
    GraphicView* iview = vs->GetView(i);
    if (iview->IsA(MAPFCLASS_VIEW)) {
      GraphicView* fview = FindInFCView((MapFClassView*)iview, e);
      if (fview) {
	view = fview;
	break;
      }
    }
    else if (iview->IsA(OVERLAYS_VIEW)) {
      GraphicView* lview = FindInList((OverlaysView*)iview, e);
      if (lview) {
	view = lview;
	break;
      }	
    }
  }
  return view;
}

GraphicView* MapInspectTool::FindInFCView(MapFClassView* vs, Event& e) {
  Selection* fs = vs->ViewIntersecting(e.x-SLOP, e.y-SLOP, e.x+SLOP, e.y+SLOP);
  if (!fs->IsEmpty()) {
    Iterator fi;
    fs->First(fi);
    return fs->GetView(fi);
  }
  else
    return nil;
} 

GraphicView* MapInspectTool::FindFeature(OverlaysView* sel, Event& e,
				     Selection* vs) {
    GraphicView* view = nil;
#if 0
    Iterator i;
    for (sel->Last(i); !sel->Done(i); sel->Prev(i)) {
	GraphicView* iview = sel->GetView(i);
	if (iview->IsA(MAPDB_VIEW)) {
	    Iterator j;
	    for (iview->Last(j); !iview->Done(j); iview->Prev(j)) {
		GraphicView* jview = iview->GetView(j); // libview
		Iterator k;
		for (jview->Last(k); !jview->Done(k); jview->Prev(k)) {
		    GraphicView* kview = jview->GetView(k); // covview
		    Iterator g;
		    for (kview->Last(g); !kview->Done(g); kview->Prev(g)) {
			MapFClassView* fcview = (MapFClassView*)kview->GetView(g);
			Selection* fs = fcview->ViewIntersecting(e.x-SLOP, e.y-SLOP, e.x+SLOP, e.y+SLOP);
			if (!fs->IsEmpty()) {
			    Iterator f;
			    fs->First(f);
			    view = fs->GetView(f);
			    vs->Exclusive(fs);
			    delete fs;
			    break;
			}
		    }
		    if (view)
			break;
		}
		if (view)
		    break;
	    }
	    break;
	} else if (iview->IsA(MAPFCLASS_VIEW)) {
	  Selection* fs = iview->ViewIntersecting(e.x-SLOP, e.y-SLOP, e.x+SLOP, e.y+SLOP);
	  if (!fs->IsEmpty()) {
	    Iterator f;
	    fs->First(f);
	    view = fs->GetView(f);
	    vs->Exclusive(fs);
	    delete fs;
	    Selection* ns = new Selection();

	    // test road network connectivity by highlighting connected
	    // features
	    if (view->IsA(MAPFEATURE_VIEW)) {
	      MapFeatureComp* mfcomp = (MapFeatureComp*)view->GetGraphicComp();
	      MapFeature* mf = mfcomp->feature();
	      if (mf && mf->type() == MapFeature::LineType) {
		MapEdge* fedge = mf->edge_primitive();
		if (fedge) {
		  if (fedge->topo()->start_node()) {
		    MapNode* snode = fedge->start_node();
		    if (snode && snode->feature()
			&& snode->feature()->graphic()) {
		      MapFeatureComp* scomp = (MapFeatureComp*)
			snode->feature()->graphic();
		      MapFeatureView* sview = (MapFeatureView*)
			scomp->FindView(view->GetViewer());
		      if (sview && sview != view && !ns->Includes(sview))
			ns->Append(sview);
		      // now add segments connected to the node
		      TopoNode* tnode = snode->topo();
		      Iterator it;
		      for (tnode->first(it); !tnode->done(it); tnode->next(it)) {
			TopoEdge* tedge = tnode->edge(tnode->elem(it));
			MapEdge* medge = (MapEdge*)tedge->value();
			MapFeature* ef = medge->feature();
			MapFeatureComp* efc = (MapFeatureComp*)ef->graphic();
			MapFeatureView* efv = (MapFeatureView*)
			  efc->FindView(view->GetViewer());
			if (efv && efv != view && !ns->Includes(efv))
			  ns->Append(efv);
		      }
		    }
		  }
		  if (fedge->topo()->end_node()) {
		    MapNode* enode = fedge->end_node();
		    if (enode && enode->feature()
			&& enode->feature()->graphic()) {
		      MapFeatureComp* ecomp = (MapFeatureComp*)
			enode->feature()->graphic();
		      MapFeatureView* eview = (MapFeatureView*)
			ecomp->FindView(view->GetViewer());
		      if (eview && eview != view && !ns->Includes(eview))
			ns->Append(eview);
		      // now add segments connected to the node
		      TopoNode* tnode = enode->topo();
		      Iterator it;
		      for (tnode->first(it); !tnode->done(it); tnode->next(it)) {
			TopoEdge* tedge = tnode->edge(tnode->elem(it));
			MapEdge* medge = (MapEdge*)tedge->value();
			MapFeature* ef = medge->feature();
			MapFeatureComp* efc = (MapFeatureComp*)ef->graphic();
			MapFeatureView* efv = (MapFeatureView*)
			  efc->FindView(view->GetViewer());
			if (efv && efv != view && !ns->Includes(efv))
			  ns->Append(efv);
		      }
		    }
		  }
		}
	      }
	      else if (mf && mf->type() == MapFeature::JunctionType) {
		MapNode* fnode = mf->node_primitive();
		if (fnode) {
		  TopoNode* tnode = fnode->topo();
		  Iterator it;
		  for (tnode->first(it); !tnode->done(it); tnode->next(it)) {
		    TopoEdge* tedge = tnode->edge(tnode->elem(it));
		    MapEdge* medge = (MapEdge*)tedge->value();
		    MapFeature* ef = medge->feature();
		    MapFeatureComp* efc = (MapFeatureComp*)ef->graphic();
		    MapFeatureView* efv = (MapFeatureView*)
		      efc->FindView(view->GetViewer());
		    if (efv && efv != view && !ns->Includes(efv))
		      ns->Append(efv);
		  }
		}
	      }
	    }
	    vs->Merge(ns);
	    delete ns;

	    break;
	  }
	}
    }
#endif
    view = FindInList(sel, e);

    if (view) {
      vs->Clear();
      vs->Append(view);

	    Selection* ns = new Selection();

	    // test road network connectivity by highlighting connected
	    // features
	    if (view->IsA(MAPFEATURE_VIEW)) {
	      MapFeatureComp* mfcomp = (MapFeatureComp*)view->GetGraphicComp();
	      MapFeature* mf = mfcomp->feature();
	      if (mf && mf->type() == MapFeature::LineType) {
		MapEdge* fedge = mf->edge_primitive();
		if (fedge) {
		  if (fedge->topo()->start_node()) {
		    MapNode* snode = fedge->start_node();
		    if (snode && snode->feature()
			&& snode->feature()->graphic()) {
		      MapFeatureComp* scomp = (MapFeatureComp*)
			snode->feature()->graphic();
		      MapFeatureView* sview = (MapFeatureView*)
			scomp->FindView(view->GetViewer());
		      if (sview && sview != view && !ns->Includes(sview))
			ns->Append(sview);
		      // now add segments connected to the node
		      TopoNode* tnode = snode->topo();
		      Iterator it;
		      for (tnode->first(it); !tnode->done(it); tnode->next(it)) {
			TopoEdge* tedge = tnode->edge(tnode->elem(it));
			MapEdge* medge = (MapEdge*)tedge->value();
			MapFeature* ef = medge->feature();
			MapFeatureComp* efc = (MapFeatureComp*)ef->graphic();
			MapFeatureView* efv = (MapFeatureView*)
			  efc->FindView(view->GetViewer());
			if (efv && efv != view && !ns->Includes(efv))
			  ns->Append(efv);
		      }
		    }
		  }
		  if (fedge->topo()->end_node()) {
		    MapNode* enode = fedge->end_node();
		    if (enode && enode->feature()
			&& enode->feature()->graphic()) {
		      MapFeatureComp* ecomp = (MapFeatureComp*)
			enode->feature()->graphic();
		      MapFeatureView* eview = (MapFeatureView*)
			ecomp->FindView(view->GetViewer());
		      if (eview && eview != view && !ns->Includes(eview))
			ns->Append(eview);
		      // now add segments connected to the node
		      TopoNode* tnode = enode->topo();
		      Iterator it;
		      for (tnode->first(it); !tnode->done(it); tnode->next(it)) {
			TopoEdge* tedge = tnode->edge(tnode->elem(it));
			MapEdge* medge = (MapEdge*)tedge->value();
			MapFeature* ef = medge->feature();
			MapFeatureComp* efc = (MapFeatureComp*)ef->graphic();
			MapFeatureView* efv = (MapFeatureView*)
			  efc->FindView(view->GetViewer());
			if (efv && efv != view && !ns->Includes(efv))
			  ns->Append(efv);
		      }
		    }
		  }
		}
	      }
	      else if (mf && mf->type() == MapFeature::JunctionType) {
		MapNode* fnode = mf->node_primitive();
		if (fnode) {
		  TopoNode* tnode = fnode->topo();
		  Iterator it;
		  for (tnode->first(it); !tnode->done(it); tnode->next(it)) {
		    TopoEdge* tedge = tnode->edge(tnode->elem(it));
		    MapEdge* medge = (MapEdge*)tedge->value();
		    MapFeature* ef = medge->feature();
		    MapFeatureComp* efc = (MapFeatureComp*)ef->graphic();
		    MapFeatureView* efv = (MapFeatureView*)
		      efc->FindView(view->GetViewer());
		    if (efv && efv != view && !ns->Includes(efv))
		      ns->Append(efv);
		  }
		}
	      }
	    }
	    vs->Merge(ns);
	    delete ns;

    }

    if (!view) {
	vs->Clear();
	unidraw->Update();
    }
    return view;
}

/*****************************************************************************/

ClassId MapSelectTool::GetClassId () { return SELECT_TOOL; }

boolean MapSelectTool::IsA (ClassId id) {
    return SELECT_TOOL == id || Tool::IsA(id);
}

MapSelectTool::MapSelectTool (ControlInfo* m) : Tool(m) { }
Tool* MapSelectTool::Copy () { return new MapSelectTool(CopyControlInfo()); }

Manipulator* MapSelectTool::CreateManipulator (
    Viewer* v, Event& e, Transformer* rel
) {
    Manipulator* m = nil;
    GraphicView* views = v->GetGraphicView();
    Selection* s = v->GetSelection(), *newSel;

    newSel = views->ViewIntersecting(e.x-SLOP, e.y-SLOP, e.x+SLOP, e.y+SLOP);
    if (e.shift) {
        Localize(s, v);
    } else {
	s->Clear();
    }
    m = new DragManip(v, new RubberRect(nil,nil, e.x,e.y,e.x,e.y), rel);

    delete newSel;
    return m;
}

Command* MapSelectTool::InterpretManipulator (Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    Viewer* viewer = dm->GetViewer();
    GraphicView* views = viewer->GetGraphicView();
    Selection* s = viewer->GetSelection();
    RubberRect* rr = (RubberRect*) dm->GetRubberband();
    Selection* newSel;
    Coord l, b, r, t;

    viewer->GetEditor()->GetWindow()->cursor(hourglass);

    rr->GetCurrent(l, b, r, t);
    newSel = SelectMapFeatures(views, l, b, r, t);

    s->Exclusive(newSel);
    delete newSel;

    viewer->GetEditor()->GetWindow()->cursor(arrow);

    return nil;
}

void MapSelectTool::SelectFeatures(Selection* sel, GraphicView* views,
				   Coord l, Coord b, Coord r, Coord t)
{
  if (views->IsA(MAPFCLASS_VIEW)) {
    MapFClassView* fcview = (MapFClassView*)views;
    Selection* fs = fcview->ViewsIntersecting(l, b, r, t);
    if (!fs->IsEmpty())
      sel->Merge(fs);
  }
  else if (views->IsA(OVERLAYS_VIEW)) {
    Iterator i;
    for (views->First(i); !views->Done(i); views->Next(i)) {
      GraphicView* v = views->GetView(i);
      SelectFeatures(sel, v, l, b, r, t);
    }
  }
}

Selection* MapSelectTool::SelectMapFeatures(GraphicView* views, Coord l,
					    Coord b, Coord r, Coord t)
{
  Selection* sel = new Selection();
  SelectFeatures(sel, views, l, b, r, t);
  return sel;
}

void MapSelectTool::Localize (Selection* s, Viewer* v) {
    Iterator i;

    for (s->First(i); !s->Done(i);) {
        GraphicView* view = s->GetView(i);

        if (view->GetViewer() != v) {
            s->Remove(i);
            view->EraseHandles();

        } else {
            s->Next(i);
        }
    }
}
